"""{{ cookiecutter.project_name }} file for ensuring the package is executable
as `{{ cookiecutter.repo_name }}` and `python -m {{ cookiecutter.python_package }}`
"""
import sys
from pathlib import Path
from typing import Any

from kedro.framework.cli.utils import find_run_command
from kedro.framework.project import configure_project


def main(*args, **kwargs) -> Any:
    package_name = Path(__file__).parent.name
    configure_project(package_name)

    interactive = hasattr(sys, 'ps1')
    kwargs["standalone_mode"] = not interactive

    run = find_run_command(package_name)
    return run(*args, **kwargs)


if __name__ == "__main__":
    main()



"""``kedro.pipeline`` provides functionality to define and execute
data-driven pipelines.
"""

from .node import GroupedNodes, Node, node
from .pipeline import Pipeline, pipeline

__all__ = ["node", "pipeline", "Node", "Pipeline", "GroupedNodes"]



"""``AbstractRunner`` is the base class for all ``Pipeline`` runner
implementations.
"""

from __future__ import annotations

import logging
import os
import sys
from abc import ABC, abstractmethod
from collections import Counter, deque
from concurrent.futures import (
    FIRST_COMPLETED,
    Executor,
    Future,
    ProcessPoolExecutor,
    wait,
)
from itertools import chain
from time import perf_counter
from typing import TYPE_CHECKING, Any

from kedro.framework.hooks.manager import _NullPluginManager
from kedro.pipeline import Pipeline
from kedro.runner.task import Task

# see https://github.com/python/cpython/blob/master/Lib/concurrent/futures/process.py#L114
_MAX_WINDOWS_WORKERS = 61

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable

    from pluggy import PluginManager

    from kedro.io import CatalogProtocol, SharedMemoryCatalogProtocol
    from kedro.pipeline.node import Node


class AbstractRunner(ABC):
    """``AbstractRunner`` is the base class for all ``Pipeline`` runner
    implementations.
    """

    def __init__(
        self,
        is_async: bool = False,
    ):
        """Instantiates the runner class.

        Args:
            is_async: If True, the node inputs and outputs are loaded and saved
                asynchronously with threads. Defaults to False.
        """
        self._is_async = is_async

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(self.__module__)

    def run(
        self,
        pipeline: Pipeline,
        catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
        only_missing_outputs: bool = False,
    ) -> dict[str, Any]:
        """Run the ``Pipeline`` using the datasets provided by ``catalog``
        and save results back to the same objects.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: An implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` from which to fetch data.
            hook_manager: The ``PluginManager`` to activate hooks.
            run_id: The id of the run.
            only_missing_outputs: Run only nodes with missing outputs.

        Raises:
            ValueError: Raised when ``Pipeline`` inputs cannot be satisfied.

        Returns:
            Dictionary with pipeline outputs, where keys are dataset names
            and values are dataset objects.
        """
        # Apply missing outputs filtering if requested
        if only_missing_outputs:
            pipeline = self._filter_pipeline_for_missing_outputs(pipeline, catalog)

        # Check which datasets used in the pipeline are in the catalog or match
        # a pattern in the catalog, not including extra dataset patterns
        # Run a warm-up to materialize all datasets in the catalog before run
        warmed_up_ds = []
        for ds in pipeline.datasets():
            if ds in catalog:
                warmed_up_ds.append(ds)
            _ = catalog.get(ds, fallback_to_runtime_pattern=True)

        # Check if there are any input datasets that aren't in the catalog and
        # don't match a pattern in the catalog.
        unsatisfied = pipeline.inputs() - set(warmed_up_ds)

        if unsatisfied:
            raise ValueError(
                f"Pipeline input(s) {unsatisfied} not found in the {catalog.__class__.__name__}"
            )

        hook_or_null_manager = hook_manager or _NullPluginManager()

        if self._is_async:
            self._logger.info(
                "Asynchronous mode is enabled for loading and saving data."
            )

        start_time = perf_counter()
        self._run(pipeline, catalog, hook_or_null_manager, run_id)  # type: ignore[arg-type]
        end_time = perf_counter()
        run_duration = end_time - start_time

        self._logger.info(
            f"Pipeline execution completed successfully in {run_duration:.1f} sec."
        )

        # Now we return all pipeline outputs, but we do not load datasets data
        run_output = {ds_name: catalog[ds_name] for ds_name in pipeline.outputs()}

        return run_output

    def _filter_pipeline_for_missing_outputs(
        self, pipeline: Pipeline, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
    ) -> Pipeline:
        """Filter pipeline to only include nodes that need to run.

        Uses reverse topological order to determine which nodes need to run
        """
        original_node_count = len(pipeline.nodes)

        # Get nodes in reverse topological order
        sorted_nodes = list(pipeline.nodes)
        sorted_nodes.reverse()

        # Build node mapping
        node_children = _build_node_children_map(pipeline)

        # Determine which nodes need to run
        nodes_to_run = _determine_nodes_to_run(
            sorted_nodes, catalog, node_children, self._logger
        )

        # Create filtered pipeline
        if not nodes_to_run:
            self._log_filtering_results(
                original_node_count, Pipeline([]), pipeline.nodes
            )
            return Pipeline([])

        filtered_pipeline = pipeline.filter(node_names=[n.name for n in nodes_to_run])
        self._log_filtering_results(
            original_node_count, filtered_pipeline, pipeline.nodes
        )

        return filtered_pipeline

    def _log_filtering_results(
        self,
        original_node_count: int,
        filtered_pipeline: Pipeline,
        all_nodes: list[Node],
    ) -> None:
        """Log the results of pipeline filtering."""
        final_node_count = len(filtered_pipeline.nodes)

        if final_node_count == 0:
            self._logger.info(
                f"Skipping all {original_node_count} nodes (all persistent outputs exist)"
            )
            return

        skipped_count = original_node_count - final_node_count
        if skipped_count > 0:
            all_node_names = {n.name for n in all_nodes}
            running_node_names = {n.name for n in filtered_pipeline.nodes}
            skipped_names = all_node_names - running_node_names
            self._logger.info(
                f"Skipping {skipped_count} nodes with existing outputs: "
                f"{', '.join(sorted(skipped_names))}"
            )

        self._logger.info(
            f"Running {final_node_count} out of {original_node_count} nodes"
        )

    @abstractmethod  # pragma: no cover
    def _get_executor(self, max_workers: int) -> Executor | None:
        """Abstract method to provide the correct executor (e.g., ThreadPoolExecutor, ProcessPoolExecutor or None if running sequentially)."""
        pass

    @abstractmethod  # pragma: no cover
    def _run(
        self,
        pipeline: Pipeline,
        catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
    ) -> None:
        """The abstract interface for running pipelines, assuming that the
         inputs have already been checked and normalized by run().
         This contains the Common pipeline execution logic using an executor.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: An implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` from which to fetch data.
            hook_manager: The ``PluginManager`` to activate hooks.
            run_id: The id of the run.
        """

        nodes = pipeline.nodes

        self._validate_catalog(catalog)
        self._validate_nodes(nodes)
        self._set_manager_datasets(catalog)

        load_counts = Counter(chain.from_iterable(n.inputs for n in pipeline.nodes))
        node_dependencies = pipeline.node_dependencies
        todo_nodes = set(node_dependencies.keys())
        done_nodes: set[Node] = set()
        futures = set()
        done = None
        max_workers = self._get_required_workers_count(pipeline)

        pool = self._get_executor(max_workers)
        if pool is None:
            for exec_index, node in enumerate(nodes):
                try:
                    Task(
                        node=node,
                        catalog=catalog,
                        hook_manager=hook_manager,
                        is_async=self._is_async,
                        run_id=run_id,
                    ).execute()
                    done_nodes.add(node)
                except Exception:
                    self._suggest_resume_scenario(pipeline, done_nodes, catalog)
                    raise
                self._logger.info("Completed node: %s", node.name)
                self._logger.info(
                    "Completed %d out of %d tasks", len(done_nodes), len(nodes)
                )
                self._release_datasets(node, catalog, load_counts, pipeline)

            return  # Exit early since everything runs sequentially

        with pool as executor:
            while True:
                ready = {n for n in todo_nodes if node_dependencies[n] <= done_nodes}
                todo_nodes -= ready
                for node in ready:
                    task = Task(
                        node=node,
                        catalog=catalog,
                        hook_manager=hook_manager,
                        is_async=self._is_async,
                        run_id=run_id,
                    )
                    if isinstance(executor, ProcessPoolExecutor):
                        task.parallel = True
                    futures.add(executor.submit(task))
                if not futures:
                    if todo_nodes:
                        self._raise_runtime_error(todo_nodes, done_nodes, ready, done)
                    break
                done, futures = wait(futures, return_when=FIRST_COMPLETED)
                for future in done:
                    try:
                        node = future.result()
                    except Exception:
                        self._suggest_resume_scenario(pipeline, done_nodes, catalog)
                        raise
                    done_nodes.add(node)
                    self._logger.info("Completed node: %s", node.name)
                    self._logger.info(
                        "Completed %d out of %d tasks", len(done_nodes), len(nodes)
                    )
                    self._release_datasets(node, catalog, load_counts, pipeline)

    @staticmethod
    def _raise_runtime_error(
        todo_nodes: set[Node],
        done_nodes: set[Node],
        ready: set[Node],
        done: set[Future[Node]] | None,
    ) -> None:
        debug_data = {
            "todo_nodes": todo_nodes,
            "done_nodes": done_nodes,
            "ready_nodes": ready,
            "done_futures": done,
        }
        debug_data_str = "\n".join(f"{k} = {v}" for k, v in debug_data.items())
        raise RuntimeError(
            f"Unable to schedule new tasks although some nodes "
            f"have not been run:\n{debug_data_str}"
        )

    def _suggest_resume_scenario(
        self,
        pipeline: Pipeline,
        done_nodes: Iterable[Node],
        catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
    ) -> None:
        """
        Suggest a command to the user to resume a run after it fails.
        The run should be started from the point closest to the failure
        for which persisted input exists.

        Args:
            pipeline: the ``Pipeline`` of the run.
            done_nodes: the ``Node``s that executed successfully.
            catalog: an implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` of the run.

        """
        remaining_nodes = set(pipeline.nodes) - set(done_nodes)

        postfix = ""
        if done_nodes:
            start_node_names = _find_nodes_to_resume_from(
                pipeline=pipeline,
                unfinished_nodes=remaining_nodes,
                catalog=catalog,
            )
            start_nodes_str = ",".join(sorted(start_node_names))
            postfix += f'  --from-nodes "{start_nodes_str}"'

        if not postfix:
            self._logger.warning(
                "No nodes ran. Repeat the previous command to attempt a new run."
            )
        else:
            self._logger.warning(
                f"There are {len(remaining_nodes)} nodes that have not run.\n"
                "You can resume the pipeline run from the nearest nodes with "
                "persisted inputs by adding the following "
                f"argument to your previous command:\n{postfix}"
            )

    @staticmethod
    def _release_datasets(
        node: Node,
        catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
        load_counts: dict,
        pipeline: Pipeline,
    ) -> None:
        """Decrement dataset load counts and release any datasets we've finished with"""
        for dataset in node.inputs:
            load_counts[dataset] -= 1
            if load_counts[dataset] < 1 and dataset not in pipeline.inputs():
                catalog.release(dataset)
        for dataset in node.outputs:
            if load_counts[dataset] < 1 and dataset not in pipeline.outputs():
                catalog.release(dataset)

    def _validate_catalog(
        self, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
    ) -> None:
        # Add catalog validation logic here if needed
        pass

    def _validate_nodes(self, node: Iterable[Node]) -> None:
        # Add node validation logic here if needed
        pass

    def _set_manager_datasets(
        self, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
    ) -> None:
        # Set up any necessary manager datasets here
        pass

    def _get_required_workers_count(self, pipeline: Pipeline) -> int:
        return 1

    @classmethod
    def _validate_max_workers(cls, max_workers: int | None) -> int:
        """
        Validates and returns the number of workers. Sets to os.cpu_count() or 1 if max_workers is None,
        and limits max_workers to 61 on Windows.

        Args:
            max_workers: Desired number of workers. If None, defaults to os.cpu_count() or 1.

        Returns:
            A valid number of workers to use.

        Raises:
            ValueError: If max_workers is set and is not positive.
        """
        if max_workers is None:
            max_workers = os.cpu_count() or 1
            if sys.platform == "win32":
                max_workers = min(_MAX_WINDOWS_WORKERS, max_workers)
        elif max_workers <= 0:
            raise ValueError("max_workers should be positive")

        return max_workers


def _find_nodes_to_resume_from(
    pipeline: Pipeline,
    unfinished_nodes: Collection[Node],
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
) -> set[str]:
    """Given a collection of unfinished nodes in a pipeline using
    a certain catalog, find the node names to pass to pipeline.from_nodes()
    to cover all unfinished nodes, including any additional nodes
    that should be re-run if their outputs are not persisted.

    Args:
        pipeline: the ``Pipeline`` to find starting nodes for.
        unfinished_nodes: collection of ``Node``s that have not finished yet
        catalog: an implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` of the run.

    Returns:
        Set of node names to pass to pipeline.from_nodes() to continue
        the run.

    """
    nodes_to_be_run = _find_all_nodes_for_resumed_pipeline(
        pipeline, unfinished_nodes, catalog
    )

    # Find which of the remaining nodes would need to run first (in topo sort)
    persistent_ancestors = _find_initial_node_group(pipeline, nodes_to_be_run)

    return {n.name for n in persistent_ancestors}


def _find_all_nodes_for_resumed_pipeline(
    pipeline: Pipeline,
    unfinished_nodes: Iterable[Node],
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
) -> set[Node]:
    """Breadth-first search approach to finding the complete set of
    ``Node``s which need to run to cover all unfinished nodes,
    including any additional nodes that should be re-run if their outputs
    are not persisted.

    Args:
        pipeline: the ``Pipeline`` to analyze.
        unfinished_nodes: the iterable of ``Node``s which have not finished yet.
        catalog: an implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` of the run.

    Returns:
        A set containing all input unfinished ``Node``s and all remaining
        ``Node``s that need to run in case their outputs are not persisted.

    """
    nodes_to_run = set(unfinished_nodes)
    initial_nodes = _nodes_with_external_inputs(unfinished_nodes)

    queue, visited = deque(initial_nodes), set(initial_nodes)
    while queue:
        current_node = queue.popleft()
        nodes_to_run.add(current_node)
        # Look for parent nodes which produce non-persistent inputs (if those exist)
        non_persistent_inputs = _enumerate_non_persistent_inputs(current_node, catalog)
        for node in _enumerate_nodes_with_outputs(pipeline, non_persistent_inputs):
            if node in visited:
                continue
            visited.add(node)
            queue.append(node)

    # Make sure no downstream tasks are skipped
    nodes_to_run = set(pipeline.from_nodes(*(n.name for n in nodes_to_run)).nodes)

    return nodes_to_run


def _nodes_with_external_inputs(nodes_of_interest: Iterable[Node]) -> set[Node]:
    """For given ``Node``s , find their subset which depends on
    external inputs of the ``Pipeline`` they constitute. External inputs
    are pipeline inputs not produced by other ``Node``s in the ``Pipeline``.

    Args:
        nodes_of_interest: the ``Node``s to analyze.

    Returns:
        A set of ``Node``s that depend on external inputs
        of nodes of interest.

    """
    p_nodes_of_interest = Pipeline(nodes_of_interest)
    p_nodes_with_external_inputs = p_nodes_of_interest.only_nodes_with_inputs(
        *p_nodes_of_interest.inputs()
    )
    return set(p_nodes_with_external_inputs.nodes)


def _enumerate_non_persistent_inputs(
    node: Node, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
) -> set[str]:
    """Enumerate non-persistent input datasets of a ``Node``.

    Args:
        node: the ``Node`` to check the inputs of.
        catalog: an implemented instance of ``CatalogProtocol`` or ``SharedMemoryCatalogProtocol`` of the run.

    Returns:
        Set of names of non-persistent inputs of given ``Node``.

    """
    # We use _datasets because they pertain parameter name format
    catalog_datasets = catalog._datasets
    non_persistent_inputs: set[str] = set()
    for node_input in node.inputs:
        if node_input.startswith("params:"):
            continue

        if (
            node_input not in catalog_datasets
            or catalog_datasets[node_input]._EPHEMERAL
        ):
            non_persistent_inputs.add(node_input)

    return non_persistent_inputs


def _enumerate_nodes_with_outputs(
    pipeline: Pipeline, outputs: Collection[str]
) -> list[Node]:
    """For given outputs, returns a list containing nodes that
    generate them in the given ``Pipeline``.

    Args:
        pipeline: the ``Pipeline`` to search for nodes in.
        outputs: the dataset names to find source nodes for.

    Returns:
        A list of all ``Node``s that are producing ``outputs``.

    """
    parent_pipeline = pipeline.only_nodes_with_outputs(*outputs)
    return parent_pipeline.nodes


def _find_initial_node_group(pipeline: Pipeline, nodes: Iterable[Node]) -> list[Node]:
    """Given a collection of ``Node``s in a ``Pipeline``,
    find the initial group of ``Node``s to be run (in topological order).

    This can be used to define a sub-pipeline with the smallest possible
    set of nodes to pass to --from-nodes.

    Args:
        pipeline: the ``Pipeline`` to search for initial ``Node``s in.
        nodes: the ``Node``s to find initial group for.

    Returns:
        A list of initial ``Node``s to run given inputs (in topological order).

    """
    node_names = {n.name for n in nodes}
    if len(node_names) == 0:
        return []
    sub_pipeline = pipeline.only_nodes(*node_names)
    initial_nodes = sub_pipeline.grouped_nodes[0]
    return initial_nodes


def _build_node_children_map(pipeline: Pipeline) -> dict[Node, set[Node]]:
    """Build mapping of nodes to their children.

    Returns:
        Dictionary mapping each node to set of its child nodes.
    """
    # Empty children sets
    node_children: dict[Node, set[Node]] = {node: set() for node in pipeline.nodes}

    # Build map for efficient lookup
    output_to_node: dict[str, Node] = {}
    for node in pipeline.nodes:
        for output in node.outputs:
            output_to_node[output] = node

    # Build children relationships
    for node in pipeline.nodes:
        for input_dataset in node.inputs:
            if input_dataset in output_to_node:
                parent = output_to_node[input_dataset]
                node_children[parent].add(node)

    return node_children


def _determine_nodes_to_run(
    sorted_nodes: list[Node],
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
    node_children: dict[Node, set[Node]],
    logger: logging.Logger,
) -> set[Node]:
    """Determine which nodes need to run based on missing outputs logic."""
    nodes_to_run: set[Node] = set()

    for node in sorted_nodes:
        if _should_node_run(node, catalog, nodes_to_run, node_children, logger):
            nodes_to_run.add(node)

    return nodes_to_run


def _should_node_run(
    node: Node,
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
    nodes_to_run: set[Node],
    node_children: dict[Node, set[Node]],
    logger: logging.Logger,
) -> bool:
    """Check if a node should run based on following rules:
    1. Always run nodes with no outputs
    2. Run if node has missing persistent outputs
    3. Run if node's outputs are needed by children that will run
    """
    # Always run nodes with no outputs
    if not node.outputs:
        logger.debug(f"Node '{node.name}' must run: has no outputs")
        return True

    # Run if node has missing persistent outputs
    if _has_missing_persistent_outputs(node, catalog, logger):
        return True

    # Run if node's outputs are needed by children that will run
    if _outputs_needed_by_children(node, catalog, nodes_to_run, node_children, logger):
        return True

    return False


def _has_missing_persistent_outputs(
    node: Node,
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
    logger: logging.Logger,
) -> bool:
    """Check if node has any persistent outputs that don't exist."""
    for output in node.outputs:
        if _is_persistent_dataset_missing(output, catalog):
            logger.debug(f"Node '{node.name}' must run: has missing output '{output}'")
            return True
    return False


def _outputs_needed_by_children(
    node: Node,
    catalog: CatalogProtocol | SharedMemoryCatalogProtocol,
    nodes_to_run: set[Node],
    node_children: dict[Node, set[Node]],
    logger: logging.Logger,
) -> bool:
    """Check if any of node's outputs are needed by children that will run."""
    children = node_children.get(node, set())

    # Only check children that are scheduled to run
    running_children = children & nodes_to_run

    for child in running_children:
        # Find datasets that connect this node to the child
        shared_datasets = set(node.outputs) & set(child.inputs)

        for dataset in shared_datasets:
            if _is_dataset_ephemeral_or_missing(dataset, catalog):
                logger.debug(
                    f"Node '{node.name}' must run: produces '{dataset}' "
                    f"needed by running child '{child.name}'"
                )
                return True

    return False


def _is_persistent_dataset_missing(
    dataset_name: str, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
) -> bool:
    """Check if a dataset is persistent and doesn't exist.

    Returns False for ephemeral datasets or existing persistent datasets.
    """
    # Not in catalog will be MemoryDataset
    if dataset_name not in catalog:
        return False

    dataset = catalog.get(dataset_name)

    # Ephemeral datasets aren't considered "missing"
    if getattr(dataset, "_EPHEMERAL", False):
        return False

    # Check if persistent dataset exists
    return not catalog.exists(dataset_name)


def _is_dataset_ephemeral_or_missing(
    dataset_name: str, catalog: CatalogProtocol | SharedMemoryCatalogProtocol
) -> bool:
    """Check if a dataset is ephemeral or missing.

    This is used to determine if a parent node needs to run to provide
    data for a child node.
    """
    # Not in catalog it will be MemoryDataset
    if dataset_name not in catalog:
        return True

    dataset = catalog.get(dataset_name)

    # Ephemeral datasets always need to be made
    if getattr(dataset, "_EPHEMERAL", False):
        return True

    # Persistent datasets need to be made if they don't exist
    return not catalog.exists(dataset_name)



"""Test a set of helper functions being used across kedro components."""

from typing import Any, NoReturn, TypeVar

import pytest

from kedro.utils import load_obj

T = TypeVar("T")


class DummyClass:
    pass


class TestExtractObject:
    def test_load_obj(self):
        extracted_obj = load_obj("tests.test_utils.DummyClass")
        assert extracted_obj is DummyClass

    def test_load_obj_default_path(self):
        extracted_obj = load_obj("DummyClass", "tests.test_utils")
        assert extracted_obj is DummyClass

    def test_load_obj_invalid_module(self):
        with pytest.raises(ImportError, match=r"No module named 'missing_path'"):
            load_obj("InvalidClass", "missing_path")


def identity(input1) -> T:
    """Return the input unchanged."""
    return input1  # pragma: no cover


def biconcat(input1: str, input2: str) -> str:
    """Concatenate two strings."""
    return f"{input1}{input2}"  # pragma: no cover


def triconcat(input1: str, input2: str, input3: str) -> str:
    """Concatenate three strings."""
    return f"{input1}{input2}{input3}"  # pragma: no cover


def constant_output() -> str:
    """Return a constant output string."""
    return "output"  # pragma: no cover


# Runner test utilities
def source() -> str:
    """Return a constant source value."""
    return "stuff"


def sink(arg: Any) -> None:
    """No-op sink function."""
    pass


def fan_in(*args: Any) -> tuple[Any, ...]:
    """Return all arguments as a tuple."""
    return args


def first_arg(*args: Any) -> Any:
    """Return the first argument."""
    return args[0]


def exception_fn(*args: Any) -> NoReturn:
    """Raise a test exception."""
    raise Exception("test exception")


def return_none(arg: Any) -> None:
    """Return None."""
    return None


def return_not_serialisable(arg: Any) -> Any:
    """Return a non-serializable object (lambda)."""
    return lambda x: x


def multi_input_list_output(arg1: Any, arg2: Any, arg3: Any = None) -> list:
    """Return first two arguments as a list."""
    return [arg1, arg2]



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import importlib
import time
from pathlib import Path

import yaml

from kedro.io import DataCatalog
from kedro.io.data_catalog import SharedMemoryDataCatalog
from kedro.pipeline import node, pipeline


# Simulate an I/O-bound task
def io_bound_task(input_data):
    time.sleep(2)  # Simulate an I/O wait (e.g., reading from a file)
    output = input_data
    return output


# Simulate a compute-bound task (matrix multiplication)
def compute_bound_task(input_data) -> str:
    # Simulate heavy compute that are not using multicore (not pandas/numpy etc)
    ans = 1
    for i in range(1, 50000):
        ans = ans * i
    return "dummy"


CATALOG_CONFIG = """
'output_{pattern}':
    type: pandas.CSVDataset
    filepath: benchmarks/data/'{pattern}.csv'

'numpy_{pattern}':
    type: pickle.PickleDataset
    filepath: benchmarks/data/'{pattern}.pkl'

'{catch_all_dataset_pattern}':
    type: pandas.CSVDataset
    filepath: benchmarks/data/data.csv
"""
CATALOG_CONFIG_DATA = yaml.safe_load(CATALOG_CONFIG)


def create_io_bound_node(inputs=None, outputs=None, name=None):
    io_node = node(io_bound_task, inputs=inputs, outputs=outputs, name=name)
    return io_node


def create_io_bound_pipeline():
    dummy_pipeline = pipeline(
        [
            create_io_bound_node("dummy_1", "output_1"),
            create_io_bound_node("dummy_2", "output_2"),
            create_io_bound_node("dummy_3", "output_3"),
            create_io_bound_node("dummy_4", "output_4"),
            create_io_bound_node("dummy_5", "output_5"),
            create_io_bound_node("dummy_6", "output_6"),
            create_io_bound_node("dummy_7", "output_7"),
            create_io_bound_node("dummy_1", "output_8"),
            create_io_bound_node("dummy_1", "output_9"),
            create_io_bound_node("dummy_1", "output_10"),
        ]
    )
    return dummy_pipeline


def create_compute_bound_node(inputs=None, outputs=None, name=None):
    io_node = node(compute_bound_task, inputs=inputs, outputs=outputs, name=name)
    return io_node


def create_compute_bound_pipeline():
    dummy_pipeline = pipeline(
        [
            create_compute_bound_node("dummy_1", "numpy_1"),
            create_compute_bound_node("dummy_2", "numpy_2"),
            create_compute_bound_node("dummy_3", "numpy_3"),
            create_compute_bound_node("dummy_4", "numpy_4"),
            create_compute_bound_node("dummy_5", "numpy_5"),
            create_compute_bound_node("dummy_6", "numpy_6"),
            create_compute_bound_node("dummy_7", "numpy_7"),
            create_compute_bound_node("dummy_1", "numpy_8"),
            create_compute_bound_node("dummy_1", "numpy_9"),
            create_compute_bound_node("dummy_1", "numpy_10"),
        ]
    )
    return dummy_pipeline


def _get_catalog(runner_name):
    """Instantiate the correct catalog based on the runner type."""
    if runner_name == "ParallelRunner":
        return SharedMemoryDataCatalog.from_config(CATALOG_CONFIG_DATA)
    return DataCatalog.from_config(CATALOG_CONFIG_DATA)


class RunnerMemorySuite:
    params = (
        "SequentialRunner",
        "ThreadRunner",
        "ParallelRunner",
    )
    param_names = ("runner",)

    def setup(self, *args, **kwargs):
        data_dir = Path("benchmarks/data")
        data_dir.mkdir(exist_ok=True, parents=True)

        # Create a dummy csv
        with open(data_dir / "data.csv", "w") as f:
            f.write("col1,col2\n1,2\n")

    def mem_runners(self, runner):
        catalog = _get_catalog(runner)
        test_pipeline = create_compute_bound_pipeline()
        runner_module = importlib.import_module("kedro.runner")
        runner_obj = getattr(runner_module, runner)()
        runner_obj.run(test_pipeline, catalog=catalog)

    def peakmem_runners(self, runner):
        catalog = _get_catalog(runner)
        test_pipeline = create_compute_bound_pipeline()
        runner_module = importlib.import_module("kedro.runner")
        runner_obj = getattr(runner_module, runner)()
        runner_obj.run(test_pipeline, catalog=catalog)


class RunnerTimeSuite:
    params = (
        "SequentialRunner",
        "ThreadRunner",
        "ParallelRunner",
    )
    param_names = ("runner",)

    def setup(self, *args, **kwargs):
        data_dir = Path("benchmarks/data")
        data_dir.mkdir(exist_ok=True, parents=True)

        # Create a dummy csv
        with open(data_dir / "data.csv", "w") as f:
            f.write("col1,col2\n1,2\n")

    def time_compute_bound_runner(self, runner):
        catalog = _get_catalog(runner)
        test_pipeline = create_compute_bound_pipeline()
        runner_module = importlib.import_module("kedro.runner")
        runner_obj = getattr(runner_module, runner)()
        runner_obj.run(test_pipeline, catalog=catalog)

    def time_io_bound_runner(self, runner):
        """IO bound pipeline"""
        catalog = _get_catalog(runner)
        test_pipeline = create_io_bound_pipeline()
        runner_module = importlib.import_module("kedro.runner")
        runner_obj = getattr(runner_module, runner)()
        runner_obj.run(test_pipeline, catalog=catalog)



from __future__ import annotations

import inspect
import itertools as it
import multiprocessing
from collections.abc import Iterable, Iterator
from concurrent.futures import (
    ALL_COMPLETED,
    Future,
    ThreadPoolExecutor,
    as_completed,
    wait,
)
from typing import TYPE_CHECKING, Any

from more_itertools import interleave

from kedro.framework.hooks.manager import (
    _create_hook_manager,
    _register_hooks,
    _register_hooks_entry_points,
)
from kedro.framework.project import settings

if TYPE_CHECKING:
    from pluggy import PluginManager

    from kedro.io import CatalogProtocol
    from kedro.pipeline.node import Node


class TaskError(Exception):
    """``TaskError`` raised by ``Task``
    in case of failure of provided task arguments

    """

    pass


class Task:
    def __init__(  # noqa: PLR0913
        self,
        node: Node,
        catalog: CatalogProtocol,
        is_async: bool,
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
        parallel: bool = False,
    ):
        self.node = node
        self.catalog = catalog
        self.hook_manager = hook_manager
        self.is_async = is_async
        self.run_id = run_id
        self.parallel = parallel

    def execute(self) -> Node:
        if self.is_async and inspect.isgeneratorfunction(self.node.func):
            raise ValueError(
                f"Async data loading and saving does not work with "
                f"nodes wrapping generator functions. Please make "
                f"sure you don't use `yield` anywhere "
                f"in node {self.node!s}."
            )

        if not self.hook_manager and not self.parallel:
            raise TaskError(
                "No hook_manager provided. This is only allowed when running a ``Task`` with ``ParallelRunner``."
            )

        if self.parallel:
            from kedro.framework.project import LOGGING, PACKAGE_NAME

            hook_manager = Task._run_node_synchronization(
                package_name=PACKAGE_NAME,
                logging_config=LOGGING,  # type: ignore[arg-type]
            )
            self.hook_manager = hook_manager
        if self.is_async:
            node = self._run_node_async(
                self.node,
                self.catalog,
                self.hook_manager,  # type: ignore[arg-type]
                self.run_id,
            )
        else:
            node = self._run_node_sequential(
                self.node,
                self.catalog,
                self.hook_manager,  # type: ignore[arg-type]
                self.run_id,
            )

        for name in node.confirms:
            self.catalog.confirm(name)

        return node

    def __call__(self) -> Node:
        """Make the class instance callable by ProcessPoolExecutor."""
        return self.execute()

    @staticmethod
    def _bootstrap_subprocess(
        package_name: str, logging_config: dict[str, Any] | None = None
    ) -> None:
        from kedro.framework.project import configure_logging, configure_project

        configure_project(package_name)
        if logging_config:
            configure_logging(logging_config)

    @staticmethod
    def _run_node_synchronization(
        package_name: str | None = None,
        logging_config: dict[str, Any] | None = None,
    ) -> Any:
        """Run a single `Node` with inputs from and outputs to the `catalog`.

        A ``PluginManager`` instance is created in each subprocess because the
        ``PluginManager`` can't be serialised.

        Args:
            package_name: The name of the project Python package.
            logging_config: A dictionary containing logging configuration.

        Returns:
            The node argument.

        """
        if multiprocessing.get_start_method() == "spawn" and package_name:
            Task._bootstrap_subprocess(package_name, logging_config)

        hook_manager = _create_hook_manager()
        _register_hooks(hook_manager, settings.HOOKS)
        _register_hooks_entry_points(hook_manager, settings.DISABLE_HOOKS_FOR_PLUGINS)

        return hook_manager

    def _run_node_sequential(
        self,
        node: Node,
        catalog: CatalogProtocol,
        hook_manager: PluginManager,
        run_id: str | None = None,
    ) -> Node:
        inputs = {}

        for name in node.inputs:
            hook_manager.hook.before_dataset_loaded(dataset_name=name, node=node)
            inputs[name] = catalog.load(name)
            hook_manager.hook.after_dataset_loaded(
                dataset_name=name, data=inputs[name], node=node
            )

        is_async = False

        additional_inputs = self._collect_inputs_from_hook(
            node, catalog, inputs, is_async, hook_manager, run_id=run_id
        )
        inputs.update(additional_inputs)

        outputs = self._call_node_run(
            node, catalog, inputs, is_async, hook_manager, run_id=run_id
        )

        items: Iterable = outputs.items()
        # if all outputs are iterators, then the node is a generator node
        if all(isinstance(d, Iterator) for d in outputs.values()):
            # Python dictionaries are ordered, so we are sure
            # the keys and the chunk streams are in the same order
            # [a, b, c]
            keys = list(outputs.keys())
            # [Iterator[chunk_a], Iterator[chunk_b], Iterator[chunk_c]]
            streams = list(outputs.values())
            # zip an endless cycle of the keys
            # with an interleaved iterator of the streams
            # [(a, chunk_a), (b, chunk_b), ...] until all outputs complete
            items = zip(it.cycle(keys), interleave(*streams))

        for name, data in items:
            hook_manager.hook.before_dataset_saved(
                dataset_name=name, data=data, node=node
            )
            catalog.save(name, data)
            hook_manager.hook.after_dataset_saved(
                dataset_name=name, data=data, node=node
            )
        return node

    def _run_node_async(
        self,
        node: Node,
        catalog: CatalogProtocol,
        hook_manager: PluginManager,  # type: ignore[arg-type]
        run_id: str | None = None,
    ) -> Node:
        with ThreadPoolExecutor() as pool:
            inputs: dict[str, Future] = {}

            for name in node.inputs:
                inputs[name] = pool.submit(
                    self._synchronous_dataset_load, name, node, catalog, hook_manager
                )

            wait(inputs.values(), return_when=ALL_COMPLETED)
            inputs = {key: value.result() for key, value in inputs.items()}
            is_async = True

            additional_inputs = self._collect_inputs_from_hook(
                node, catalog, inputs, is_async, hook_manager, run_id=run_id
            )
            inputs.update(additional_inputs)

            outputs = self._call_node_run(
                node, catalog, inputs, is_async, hook_manager, run_id=run_id
            )

            future_dataset_mapping = {}
            for name, data in outputs.items():
                hook_manager.hook.before_dataset_saved(
                    dataset_name=name, data=data, node=node
                )
                future = pool.submit(catalog.save, name, data)
                future_dataset_mapping[future] = (name, data)

            for future in as_completed(future_dataset_mapping):
                exception = future.exception()
                if exception:
                    raise exception
                name, data = future_dataset_mapping[future]
                hook_manager.hook.after_dataset_saved(
                    dataset_name=name, data=data, node=node
                )
        return node

    @staticmethod
    def _synchronous_dataset_load(
        dataset_name: str,
        node: Node,
        catalog: CatalogProtocol,
        hook_manager: PluginManager,
    ) -> Any:
        """Minimal wrapper to ensure Hooks are run synchronously
        within an asynchronous dataset load."""
        hook_manager.hook.before_dataset_loaded(dataset_name=dataset_name, node=node)
        return_ds = catalog.load(dataset_name)
        hook_manager.hook.after_dataset_loaded(
            dataset_name=dataset_name, data=return_ds, node=node
        )
        return return_ds

    @staticmethod
    def _collect_inputs_from_hook(  # noqa: PLR0913
        node: Node,
        catalog: CatalogProtocol,
        inputs: dict[str, Any],
        is_async: bool,
        hook_manager: PluginManager,
        run_id: str | None = None,
    ) -> dict[str, Any]:
        inputs = (
            inputs.copy()
        )  # shallow copy to prevent in-place modification by the hook
        hook_response = hook_manager.hook.before_node_run(
            node=node,
            catalog=catalog,
            inputs=inputs,
            is_async=is_async,
            run_id=run_id,
        )

        additional_inputs = {}
        if (
            hook_response is not None
        ):  # all hooks on a _NullPluginManager will return None instead of a list
            for response in hook_response:
                if response is not None and not isinstance(response, dict):
                    response_type = type(response).__name__
                    raise TypeError(
                        f"'before_node_run' must return either None or a dictionary mapping "
                        f"dataset names to updated values, got '{response_type}' instead."
                    )
                additional_inputs.update(response or {})

        return additional_inputs

    @staticmethod
    def _call_node_run(  # noqa: PLR0913
        node: Node,
        catalog: CatalogProtocol,
        inputs: dict[str, Any],
        is_async: bool,
        hook_manager: PluginManager,
        run_id: str | None = None,
    ) -> dict[str, Any]:
        try:
            outputs = node.run(inputs)
        except Exception as exc:
            hook_manager.hook.on_node_error(
                error=exc,
                node=node,
                catalog=catalog,
                inputs=inputs,
                is_async=is_async,
                run_id=run_id,
            )
            raise exc
        hook_manager.hook.after_node_run(
            node=node,
            catalog=catalog,
            inputs=inputs,
            outputs=outputs,
            is_async=is_async,
            run_id=run_id,
        )
        return outputs



"""``SequentialRunner`` is an ``AbstractRunner`` implementation. It can be
used to run the ``Pipeline`` in a sequential manner using a topological sort
of provided nodes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from kedro.runner.runner import AbstractRunner

if TYPE_CHECKING:
    from pluggy import PluginManager

    from kedro.io import CatalogProtocol
    from kedro.pipeline import Pipeline


class SequentialRunner(AbstractRunner):
    """``SequentialRunner`` is an ``AbstractRunner`` implementation. It can
    be used to run the ``Pipeline`` in a sequential manner using a
    topological sort of provided nodes.
    """

    def __init__(
        self,
        is_async: bool = False,
    ):
        """Instantiates the runner class.

        Args:
            is_async: If True, the node inputs and outputs are loaded and saved
                asynchronously with threads. Defaults to False.
        """
        super().__init__(is_async=is_async)

    def _get_executor(self, max_workers: int) -> None:
        return None

    def _run(
        self,
        pipeline: Pipeline,
        catalog: CatalogProtocol,
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
    ) -> None:
        """The method implementing sequential pipeline running.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: An implemented instance of ``CatalogProtocol`` from which to fetch data.
            hook_manager: The ``PluginManager`` to activate hooks.
            run_id: The id of the run.

        Raises:
            Exception: in case of any downstream node failure.
        """
        if not self._is_async:
            self._logger.info(
                "Using synchronous mode for loading and saving data. Use the --async flag "
                "for potential performance gains. https://docs.kedro.org/en/stable/nodes_and_pipelines/run_a_pipeline.html#load-and-save-asynchronously"
            )
        super()._run(
            pipeline=pipeline,
            catalog=catalog,
            hook_manager=hook_manager,
            run_id=run_id,
        )



"""A collection of CLI commands for working with Kedro catalog."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import click
import yaml
from click import secho

from kedro.framework.cli.utils import env_option, split_string
from kedro.framework.session import KedroSession

if TYPE_CHECKING:
    from kedro.framework.startup import ProjectMetadata


def _create_session(package_name: str, **kwargs: Any) -> KedroSession:
    kwargs.setdefault("save_on_close", False)
    return KedroSession.create(**kwargs)


@click.group(name="kedro")
def catalog_cli() -> None:  # pragma: no cover
    pass


@catalog_cli.group()
def catalog() -> None:
    """Commands for working with catalog."""


@catalog.command("describe-datasets")
@env_option
@click.option(
    "--pipeline",
    "-p",
    type=str,
    default="",
    help="Name of the modular pipeline to run. If not set, "
    "the project pipeline is run by default.",
    callback=split_string,
)
@click.pass_obj
def describe_datasets(metadata: ProjectMetadata, pipeline: str, env: str) -> None:
    """
    Describe datasets used in the specified pipelines, grouped by type.\n

    This command provides a structured overview of datasets used in the selected pipelines,
    categorizing them into three groups:\n
    - `datasets`: Datasets explicitly defined in the catalog.\n
    - `factories`: Datasets resolved from dataset factory patterns.\n
    - `defaults`: Datasets that do not match any pattern or explicit definition.\n
    """
    session = _create_session(metadata.package_name, env=env)
    context = session.load_context()

    p = pipeline or None
    datasets_dict = context.catalog.describe_datasets(p)  # type: ignore

    secho(yaml.dump(datasets_dict))


@catalog.command("list-patterns")
@env_option
@click.pass_obj
def list_patterns(metadata: ProjectMetadata, env: str) -> None:
    """
    List all dataset factory patterns in the catalog, ranked by priority.

    This method retrieves all dataset factory patterns defined in the catalog,
    ordered by the priority in which they are matched.
    """
    session = _create_session(metadata.package_name, env=env)
    context = session.load_context()

    click.echo(yaml.dump(context.catalog.list_patterns()))  # type: ignore


@catalog.command("resolve-patterns")
@env_option
@click.option(
    "--pipeline",
    "-p",
    type=str,
    default="",
    help="Name of the modular pipeline to run. If not set, "
    "the project pipeline is run by default.",
    callback=split_string,
)
@click.pass_obj
def resolve_patterns(metadata: ProjectMetadata, pipeline: str, env: str) -> None:
    """
    Resolve dataset factory patterns against pipeline datasets.

    This method resolves dataset factory patterns for datasets used in the specified pipelines.
    It includes datasets explicitly defined in the catalog as well as those resolved
    from dataset factory patterns.
    """
    session = _create_session(metadata.package_name, env=env)
    context = session.load_context()

    p = pipeline or None
    datasets_dict = context.catalog.resolve_patterns(p)  # type: ignore

    secho(yaml.dump(datasets_dict))



"""``MemoryDataset`` is a dataset implementation which handles in-memory data."""

from __future__ import annotations

import copy
from typing import Any

from kedro.io.core import AbstractDataset, DatasetError, TCopyMode

_EMPTY = object()


class MemoryDataset(AbstractDataset):
    """``MemoryDataset`` loads and saves data from/to an in-memory
    Python object. The `_EPHEMERAL` attribute is set to True to
    indicate MemoryDataset's non-persistence.

    Example:
    ``` python
    from kedro.io import MemoryDataset
    import pandas as pd

    data = pd.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
    dataset = MemoryDataset(data=data)

    loaded_data = dataset.load()
    assert loaded_data.equals(data)

    new_data = pd.DataFrame({"col1": [1, 2], "col2": [4, 5]})
    dataset.save(new_data)
    reloaded_data = dataset.load()
    assert reloaded_data.equals(new_data)
    ```
    """

    def __init__(
        self,
        data: Any = _EMPTY,
        copy_mode: TCopyMode | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        """Creates a new instance of ``MemoryDataset`` pointing to the
        provided Python object.

        Args:
            data: Python object containing the data.
            copy_mode: The copy mode used to copy the data. Possible
                values are: "deepcopy", "copy" and "assign". If not
                provided, it is inferred based on the data type.
            metadata: Any arbitrary metadata.
                This is ignored by Kedro, but may be consumed by users or external plugins.
        """
        self._data = _EMPTY
        self._copy_mode = copy_mode
        self.metadata = metadata
        self._EPHEMERAL = True
        if data is not _EMPTY:
            self.save.__wrapped__(self, data)  # type: ignore[attr-defined]

    def load(self) -> Any:
        if self._data is _EMPTY:
            raise DatasetError("Data for MemoryDataset has not been saved yet.")

        copy_mode = self._copy_mode or _infer_copy_mode(self._data)
        data = _copy_with_mode(self._data, copy_mode=copy_mode)
        return data

    def save(self, data: Any) -> None:
        copy_mode = self._copy_mode or _infer_copy_mode(data)
        self._data = _copy_with_mode(data, copy_mode=copy_mode)

    def _exists(self) -> bool:
        return self._data is not _EMPTY

    def _release(self) -> None:
        self._data = _EMPTY

    def _describe(self) -> dict[str, Any]:
        if self._data is not _EMPTY:
            return {"data": f"<{type(self._data).__name__}>"}
        # the string representation of datasets leaves out __init__
        # arguments that are empty/None, equivalent here is _EMPTY
        return {"data": None}  # pragma: no cover


def _infer_copy_mode(data: Any) -> TCopyMode:
    """Infers the copy mode to use given the data type.

    Args:
        data: The data whose type will be used to infer the copy mode.

    Returns:
        One of "copy", "assign" or "deepcopy" as the copy mode to use.
    """
    try:
        import pandas as pd
    except ImportError:  # pragma: no cover
        pd = None  # type: ignore[assignment]  # pragma: no cover
    try:
        import numpy as np
    except ImportError:  # pragma: no cover
        np = None  # type: ignore[assignment] # pragma: no cover
    try:
        import ibis
    except ImportError:  # pragma: no cover
        ibis = None  # type: ignore[assignment] # pragma: no cover

    if pd and isinstance(data, pd.DataFrame) or np and isinstance(data, np.ndarray):
        return "copy"
    elif type(data).__name__ == "DataFrame" or ibis and isinstance(data, ibis.Table):
        return "assign"
    else:
        return "deepcopy"


def _copy_with_mode(data: Any, copy_mode: TCopyMode) -> Any:
    """Returns the copied data using the copy mode specified.
    If no copy mode is provided, then it is inferred based on the type of the data.

    Args:
        data: The data to copy.
        copy_mode: The copy mode to use, one of "deepcopy", "copy" and "assign".

    Raises:
        DatasetError: If copy_mode is specified, but isn't valid
            (i.e: not one of deepcopy, copy, assign)

    Returns:
        The data copied according to the specified copy mode.
    """
    if copy_mode == "deepcopy":
        copied_data = copy.deepcopy(data)
    elif copy_mode == "copy":
        copied_data = data.copy()
    elif copy_mode == "assign":
        copied_data = data
    else:
        raise DatasetError(
            f"Invalid copy mode: {copy_mode}. "
            f"Possible values are: deepcopy, copy, assign."
        )

    return copied_data


def _is_memory_dataset(ds_or_type: AbstractDataset | str) -> bool:
    """Check if dataset or str type provided is a MemoryDataset."""
    if isinstance(ds_or_type, MemoryDataset):
        return True
    if isinstance(ds_or_type, str):
        return ds_or_type in {"MemoryDataset", "kedro.io.memory_dataset.MemoryDataset"}

    return False



"""kedro is a CLI for managing Kedro projects.

This module implements commands available from the kedro CLI.
"""

from __future__ import annotations

import importlib
import sys
import traceback
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click

if TYPE_CHECKING:
    from collections.abc import Sequence

from kedro import __version__ as version
from kedro.framework.cli import BRIGHT_BLACK, ORANGE
from kedro.framework.cli.hooks import get_cli_hook_manager
from kedro.framework.cli.utils import (
    CONTEXT_SETTINGS,
    ENTRY_POINT_GROUPS,
    CommandCollection,
    KedroCliError,
    LazyGroup,
    _get_entry_points,
    load_entry_points,
)
from kedro.framework.project import LOGGING  # noqa: F401
from kedro.framework.startup import bootstrap_project
from kedro.utils import find_kedro_project, is_kedro_project

LOGO = rf"""
 _            _
| | _____  __| |_ __ ___
| |/ / _ \/ _` | '__/ _ \
|   <  __/ (_| | | | (_) |
|_|\_\___|\__,_|_|  \___/
v{version}
"""


@click.group(context_settings=CONTEXT_SETTINGS, name="kedro")
@click.version_option(version, "--version", "-V", help="Show version and exit")
def cli() -> None:  # pragma: no cover
    """Kedro is a CLI for creating and using Kedro projects. For more
    information, type ``kedro info``.

    NOTE: If a command from a plugin conflicts with a built-in command from Kedro,
    the command from the plugin will take precedence.

    """
    pass


@cli.command()
def info() -> None:
    """Get more information about kedro."""
    click.secho(LOGO, fg="green")
    click.echo(
        "Kedro is a Python framework for\n"
        "creating reproducible, maintainable\n"
        "and modular data science code."
    )

    plugin_versions = {}
    plugin_entry_points = defaultdict(set)
    for plugin_entry_point in ENTRY_POINT_GROUPS:
        for entry_point in _get_entry_points(plugin_entry_point):
            module_name = entry_point.module.split(".")[0]
            plugin_versions[module_name] = entry_point.dist.version
            plugin_entry_points[module_name].add(plugin_entry_point)

    click.echo()
    if plugin_versions:
        click.echo("Installed plugins:")
        for plugin_name, plugin_version in sorted(plugin_versions.items()):
            entrypoints_str = ",".join(sorted(plugin_entry_points[plugin_name]))
            click.echo(
                f"{plugin_name}: {plugin_version} (entry points:{entrypoints_str})"
            )
    else:  # pragma: no cover
        click.echo("No plugins installed")


@click.group(
    context_settings=CONTEXT_SETTINGS,
    cls=LazyGroup,
    name="kedro",
    lazy_subcommands={
        "registry": "kedro.framework.cli.registry.registry",
        "catalog": "kedro.framework.cli.catalog.catalog",
        "ipython": "kedro.framework.cli.project.ipython",
        "run": "kedro.framework.cli.project.run",
        "package": "kedro.framework.cli.project.package",
        "jupyter": "kedro.framework.cli.jupyter.jupyter",
        "pipeline": "kedro.framework.cli.pipeline.pipeline",
    },
)
def project_commands() -> None:
    pass  # pragma: no cover


@click.group(
    context_settings=CONTEXT_SETTINGS,
    name="kedro",
    cls=LazyGroup,
    lazy_subcommands={
        "new": "kedro.framework.cli.starters.new",
        "starter": "kedro.framework.cli.starters.starter",
    },
)
def global_commands() -> None:
    pass  # pragma: no cover


def _init_plugins() -> None:
    init_hooks = load_entry_points("init")
    for init_hook in init_hooks:
        init_hook()


class KedroCLI(CommandCollection):
    """A CommandCollection class to encapsulate the KedroCLI command
    loading.
    """

    def __init__(self, project_path: Path):
        self._metadata = None  # running in package mode
        if is_kedro_project(project_path):
            self._metadata = bootstrap_project(project_path)
        self._cli_hook_manager = get_cli_hook_manager()

        super().__init__(
            ("Global commands", self.global_groups),
            ("Project specific commands", self.project_groups),
        )

    def main(
        self,
        args: Any | None = None,
        prog_name: Any | None = None,
        complete_var: Any | None = None,
        standalone_mode: bool = True,
        **extra: Any,
    ) -> Any:
        if self._metadata:
            extra.update(obj=self._metadata)

        # This is how click's internals parse sys.argv, which include the command,
        # subcommand, arguments and options. click doesn't store this information anywhere
        # so we have to re-do it.
        args = sys.argv[1:] if args is None else list(args)
        self._cli_hook_manager.hook.before_command_run(
            project_metadata=self._metadata, command_args=args
        )

        try:
            super().main(
                args=args,
                prog_name=prog_name,
                complete_var=complete_var,
                standalone_mode=standalone_mode,
                **extra,
            )
        # click.core.main() method exits by default, we capture this and then
        # exit as originally intended

        except SystemExit as exc:
            self._cli_hook_manager.hook.after_command_run(
                project_metadata=self._metadata, command_args=args, exit_code=exc.code
            )

            # When CLI is run outside of a project, project_groups are not registered
            catch_exception = "click.exceptions.UsageError: No such command"
            # click convert exception handles to error message
            if catch_exception in traceback.format_exc() and not self.project_groups:
                warn = click.style(
                    "\nKedro project not found in this directory. ",
                    fg=ORANGE,
                    bold=True,
                )
                result = (
                    click.style("Project specific commands such as ")
                    + click.style("'run' ", fg="cyan")
                    + "or "
                    + click.style("'jupyter' ", fg="cyan")
                    + "are only available within a project directory."
                )
                message = warn + result
                hint = (
                    click.style(
                        "\nHint: Kedro is looking for a file called ", fg=BRIGHT_BLACK
                    )
                    + click.style("'pyproject.toml", fg="magenta")
                    + click.style(
                        ", is one present in your current working directory?",
                        fg=BRIGHT_BLACK,
                    )
                )
                click.echo(message)
                click.echo(hint)
            sys.exit(exc.code)
        except Exception:
            self._cli_hook_manager.hook.after_command_run(
                project_metadata=self._metadata, command_args=args, exit_code=1
            )
            raise

    @property
    def global_groups(self) -> Sequence[click.Group]:
        """Property which loads all global command groups from plugins and
        combines them with the built-in ones (eventually overriding the
        built-in ones if they are redefined by plugins).
        """
        return [cli, *load_entry_points("global"), global_commands]

    @property
    def project_groups(self) -> Sequence[click.Group]:
        """Property which loads all project command groups from the
        project and the plugins, then combines them with the built-in ones.
        Built-in commands can be overridden by plugins, which can be
        overridden by a custom project cli.py.
        See https://docs.kedro.org/en/stable/extend_kedro/common_use_cases.html#use-case-3-how-to-add-or-modify-cli-commands
        on how to add this.
        """
        if not self._metadata:
            return []

        plugins = load_entry_points("project")

        try:
            project_cli = importlib.import_module(f"{self._metadata.package_name}.cli")
            # fail gracefully if cli.py does not exist
        except ModuleNotFoundError:
            # return only built-in commands and commands from plugins
            # (plugins can override built-in commands)
            return [*plugins, project_commands]

        # fail badly if cli.py exists, but has no `cli` in it
        if not hasattr(project_cli, "cli"):
            raise KedroCliError(
                f"Cannot load commands from {self._metadata.package_name}.cli"
            )
        user_defined = project_cli.cli
        # return built-in commands, plugin commands and user defined commands
        # (overriding happens as follows built-in < plugins < cli.py)
        return [user_defined, *plugins, project_commands]


def main() -> None:  # pragma: no cover
    """Main entry point. Look for a ``cli.py``, and, if found, add its
    commands to `kedro`'s before invoking the CLI.
    """
    _init_plugins()
    cli_collection = KedroCLI(project_path=find_kedro_project(Path.cwd()) or Path.cwd())
    cli_collection()



"""``CatalogConfigResolver`` resolves dataset configurations and datasets'
patterns based on catalog configuration and credentials provided.
"""

from __future__ import annotations

import copy
import logging
import re
from typing import TYPE_CHECKING, Any

from parse import parse

from kedro.io.core import DatasetError

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

Patterns = dict[str, dict[str, Any]]

CREDENTIALS_KEY = "credentials"
DEFAULT_RUNTIME_PATTERN = {"{default}": {"type": "kedro.io.MemoryDataset"}}


def _fetch_credentials(credentials_name: str, credentials: dict[str, Any]) -> Any:
    """Fetch the specified credentials from the provided credentials dictionary.

    Args:
        credentials_name: Credentials name.
        credentials: A dictionary with all credentials.

    Returns:
        The set of requested credentials.

    Raises:
        KeyError: When a dataset with the given name has not yet been
            registered.

    """
    try:
        return credentials[credentials_name]
    except KeyError as exc:
        raise KeyError(
            f"Unable to find credentials '{credentials_name}': check your data "
            "catalog and credentials configuration. See "
            "https://docs.kedro.org/en/stable/data/index.html#DataCatalog "
            "for an example."
        ) from exc


def _resolve_credentials(
    config: dict[str, Any], credentials: dict[str, Any]
) -> dict[str, Any]:
    """Return the dataset configuration where credentials are resolved using
    credentials dictionary provided.

    Args:
        config: Original dataset config, which may contain unresolved credentials.
        credentials: A dictionary with all credentials.

    Returns:
        The dataset config, where all the credentials are successfully resolved.
    """
    config = copy.deepcopy(config)

    def _resolve_value(key: str, value: Any) -> Any:
        if key == CREDENTIALS_KEY and isinstance(value, str):
            return _fetch_credentials(value, credentials)
        if isinstance(value, dict):
            return {k: _resolve_value(k, v) for k, v in value.items()}
        return value

    return {k: _resolve_value(k, v) for k, v in config.items()}


class CatalogConfigResolver:
    """
    Resolves dataset configurations based on dataset factory patterns and credentials.

    The `CatalogConfigResolver` is responsible for managing dataset factory patterns, resolving
    credentials, and dynamically generating dataset configurations. It supports advanced
    features like pattern matching, user-set catch-all patterns, and runtime patterns, enabling
    flexible and reusable dataset configurations.

    Attributes:
        _default_runtime_patterns (Patterns): Runtime patterns used for resolving datasets.
        _dataset_patterns (dict): Sorted dataset factory patterns extracted from the catalog configuration.
        _user_catch_all_pattern (dict): User provided catch all pattern.
        _resolved_configs (dict): Resolved dataset configurations with credentials applied.

    Example:
    ``` python
    from kedro.io.catalog_config_resolver import CatalogConfigResolver

    config = {
        "{namespace}.int_{name}": {
            "type": "pandas.CSVDataset",
            "filepath": "{name}.csv",
            "credentials": "db_credentials",
        }
    }
    credentials = {"db_credentials": {"user": "username", "pass": "pass"}}
    resolver = CatalogConfigResolver(config=config, credentials=credentials)
    resolved_config = resolver.resolve_pattern("data.int_customers")
    print(resolved_config)
    # {'type': 'pandas.CSVDataset', 'filepath': 'customers.csv', 'credentials': {'user': 'username', 'pass': 'pass'}}
    ```
    """

    def __init__(
        self,
        config: dict[str, dict[str, Any]] | None = None,
        credentials: dict[str, dict[str, Any]] | None = None,
        default_runtime_patterns: Patterns | None = None,
    ):
        """
        Initialize the `CatalogConfigResolver`.

        Args:
            config: Dataset configurations from the catalog.
            credentials: Credentials for datasets.
            default_runtime_patterns: Runtime patterns for resolving datasets.

        Example:
        ``` python
        from kedro.io.catalog_config_resolver import CatalogConfigResolver

        config = {
            "{namespace}.int_{name}": {
                "type": "pandas.CSVDataset",
                "filepath": "{name}.csv",
            }
        }
        resolver = CatalogConfigResolver(config=config)
        resolver._dataset_patterns
        # {'{namespace}.int_{name}': {'type': 'pandas.CSVDataset', 'filepath': '{name}.csv'}}
        ```
        """
        if default_runtime_patterns is None:
            self._logger.warning(
                f"Since runtime patterns are not provided, setting "
                f"the runtime pattern to default value: {DEFAULT_RUNTIME_PATTERN}"
            )
        self._default_runtime_patterns = (
            default_runtime_patterns or DEFAULT_RUNTIME_PATTERN
        )
        self._dataset_patterns, self._user_catch_all_pattern = self._extract_patterns(
            config, credentials
        )
        self._resolved_configs = self._resolve_credentials(config, credentials)

    @property
    def config(self) -> dict[str, dict[str, Any]]:
        """
        Get the resolved dataset configurations.

        Returns:
            Resolved dataset configurations.

        Example:
        ``` python
        from kedro.io.catalog_config_resolver import CatalogConfigResolver

        config = {
            "companies": {
                "type": "pandas.CSVDataset",
                "filepath": "companies.csv",
                "credentials": "db_credentials",
            }
        }
        credentials = {"db_credentials": {"user": "username", "pass": "pass"}}
        resolver = CatalogConfigResolver(config, credentials)
        print(resolver.config)
        # {'companies': {'type': 'pandas.CSVDataset', 'filepath': 'companies.csv', 'credentials': {'user': 'username', 'pass': 'pass'}}}
        ```
        """
        return self._resolved_configs

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(__name__)

    @staticmethod
    def is_pattern(pattern: str) -> bool:
        """
        Check if a given string is a pattern. Assume that any name with '{' is a pattern.

        Args:
            pattern: The string to check.

        Returns:
            True if the string is a pattern, False otherwise.

        Example:
        ``` python
        CatalogConfigResolver.is_pattern("{namespace}.int_{name}")
        # True
        CatalogConfigResolver.is_pattern("example_dataset")
        # False
        ```
        """
        return "{" in pattern

    @staticmethod
    def _pattern_specificity(pattern: str) -> int:
        """
        Calculate the specificity of a pattern based on characters outside curly brackets.

        Args:
            pattern: The pattern to analyze.

        Returns:
            The number of characters outside curly brackets.

        Example:
        ``` python
        CatalogConfigResolver._pattern_specificity("{namespace}.int_{name}")
        # 5
        ```
        """
        # Remove all the placeholders from the pattern and count the number of remaining chars
        result = re.sub(r"\{.*?\}", "", pattern)
        return len(result)

    @classmethod
    def _sort_patterns(cls, dataset_patterns: Patterns) -> Patterns:
        """Sort a dictionary of dataset patterns according to parsing rules.

        Patterns are sorted in the following order:
        1. Decreasing specificity (number of characters outside the curly brackets)
        2. Decreasing number of placeholders (number of curly bracket pairs)
        3. Alphabetically

        Args:
            dataset_patterns: A dictionary of dataset patterns.

        Returns:
            A sorted dictionary of dataset patterns.

        Raises:
            DatasetError: If multiple catch-all patterns are found.

        Example:
        ``` python
        patterns = {
            "{namespace}.int_{name}{a}": {},
            "{namespace}.{name}": {},
            "{name}": {},
            "{namespace}.ant_{name}{c}": {},
        }
        sorted_patterns = CatalogConfigResolver._sort_patterns(patterns)
        print(sorted_patterns.keys())
        # dict_keys(['{namespace}.ant_{name}{c}', '{namespace}.int_{name}{a}', '{namespace}.{name}', '{name}'])
        ```
        """
        sorted_keys = sorted(
            dataset_patterns,
            key=lambda pattern: (
                -(cls._pattern_specificity(pattern)),
                -pattern.count("{"),
                pattern,
            ),
        )
        catch_all = [
            pattern for pattern in sorted_keys if cls._pattern_specificity(pattern) == 0
        ]
        if len(catch_all) > 1:
            raise DatasetError(
                f"Multiple catch-all patterns found in the catalog: {', '.join(catch_all)}. Only one catch-all pattern is allowed, remove the extras."
            )
        return {key: dataset_patterns[key] for key in sorted_keys}

    @classmethod
    def _validate_pattern_config(cls, ds_name: str, ds_config: dict[str, Any]) -> None:
        """Checks whether a dataset pattern configuration is valid - all
        keys used in the configuration present in the dataset pattern name.

        Args:
            ds_name: Dataset pattern name.
            ds_config: Dataset pattern configuration.

        Raises:
            DatasetError: when keys used in the configuration do not present in the dataset pattern name.

        Example:
        ``` python
        pattern = "{namespace}.int_{name}"
        config = {"filepath": "{name}.csv"}
        CatalogConfigResolver._validate_pattern_config(pattern, config)
        # No error
        pattern = "{namespace}.int_{name}"
        config = {"filepath": "{nam}.csv"}
        CatalogConfigResolver._validate_pattern_config(pattern, config)
        # DatasetError: Incorrect dataset configuration provided. Keys used in the configuration {'{nam}'} should present in
        # the dataset pattern name {namespace}.int_{name}.
        ```
        """
        # Find all occurrences of {} in the string including brackets
        search_regex = r"\{.*?\}"
        name_placeholders = set(re.findall(search_regex, ds_name))
        config_placeholders = set()

        def _traverse_config(config: Any) -> None:
            if isinstance(config, dict):
                for value in config.values():
                    _traverse_config(value)
            elif isinstance(config, (list | tuple)):
                for value in config:
                    _traverse_config(value)
            elif isinstance(config, str) and "}" in config:
                config_placeholders.update(set(re.findall(search_regex, config)))

        _traverse_config(ds_config)

        if config_placeholders - name_placeholders:
            raise DatasetError(
                f"Incorrect dataset configuration provided. "
                f"Keys used in the configuration {config_placeholders - name_placeholders} "
                f"should present in the dataset pattern name {ds_name}."
            )

    @classmethod
    def _resolve_dataset_config(
        cls,
        ds_name: str,
        pattern: str,
        config: Any,
    ) -> Any:
        """
        Resolve dataset configuration based on the provided pattern.

        Args:
            ds_name: The dataset name to resolve.
            pattern: The pattern to use for resolution.
            config: The dataset configuration.

        Returns:
            The resolved dataset configuration.

        Example:
        ``` python
        pattern = "{namespace}.int_{name}"
        config = {"filepath": "{name}.csv"}
        resolved_config = CatalogConfigResolver._resolve_dataset_config(
            "data.int_customers", pattern, config
        )
        print(resolved_config)
        # {"filepath": "customers.csv"}
        ```
        """
        resolved_vars = parse(pattern, ds_name)
        # Resolve the pattern config for the dataset
        if isinstance(config, dict):
            for key, value in config.items():
                config[key] = cls._resolve_dataset_config(ds_name, pattern, value)
        elif isinstance(config, (list | tuple)):
            config = [
                cls._resolve_dataset_config(ds_name, pattern, value) for value in config
            ]
        elif isinstance(config, str) and "}" in config:
            config = config.format_map(resolved_vars.named)
        return config

    def list_patterns(self) -> list[str]:
        """
        List all patterns available in the catalog.

        Returns:
            A list of dataset patterns.

        Example:
        ``` python
        config = {
            "{namespace}.int_{name}": {
                "type": "pandas.CSVDataset",
                "filepath": "{name}.csv",
            }
        }
        resolver = CatalogConfigResolver(config)
        print(resolver.list_patterns())
        # ['{namespace}.int_{name}', '{default}']
        ```
        """
        # User defined patterns
        patterns = list(self._dataset_patterns.keys()) + list(
            self._user_catch_all_pattern.keys()
        )

        # All patterns excluding duplicates
        for pattern in self._default_runtime_patterns.keys():
            if pattern not in patterns:
                patterns.append(pattern)

        return patterns

    @classmethod
    def _get_matches(cls, patterns: Iterable[str], ds_name: str) -> Generator[str]:
        """
        Find all patterns that match a given dataset name.

        This method iterates over a collection of patterns and checks if the given
        dataset name matches any of them using the `parse` function.

        Args:
            patterns: A collection of patterns to match against.
            ds_name: The name of the dataset to match.

        Returns:
            A generator yielding patterns that match the dataset name.

        Example:
        ``` python
        patterns = ["{namespace}.int_{name}", "{name}"]
        matches = CatalogConfigResolver._get_matches(patterns, "data.int_customers")
        print(list(matches))
        # ['{namespace}.int_{name}']
        ```
        """
        return (pattern for pattern in patterns if parse(pattern, ds_name))

    def match_dataset_pattern(self, ds_name: str) -> str | None:
        """
        Match a dataset name against dataset patterns.

        This method checks if the given dataset name matches any of the dataset
        patterns defined in the catalog. If a match is found, the first matching
        pattern is returned.

        Args:
            ds_name: The name of the dataset to match.

        Returns:
            The first matching pattern, or `None` if no match is found.

        Example:
        ``` python
        config = {
            "{namespace}.int_{name}": {"type": "pandas.CSVDataset"},
            "{name}": {"type": "MemoryDataset"},
        }
        resolver = CatalogConfigResolver(config=config)
        match = resolver.match_dataset_pattern("data.int_customers")
        print(match)
        # {namespace}.int_{name}
        ```
        """
        matches = self._get_matches(self._dataset_patterns.keys(), ds_name)
        return next(matches, None)

    def match_user_catch_all_pattern(self, ds_name: str) -> str | None:
        """
        Match a dataset name against the user-defined catch-all pattern.

        This method checks if the given dataset name matches any of the user-defined
        catch-all patterns. If a match is found, the first matching pattern is returned.

        Args:
            ds_name: The name of the dataset to match.

        Returns:
            The first matching pattern, or `None` if no match is found.

        Example:
        ``` python
        config = {"{name}": {"type": "MemoryDataset"}}
        resolver = CatalogConfigResolver(config=config)
        match = resolver.match_user_catch_all_pattern("example_dataset")
        print(match)
        # {name}
        ```
        """
        user_catch_all_pattern = set(self._user_catch_all_pattern.keys())
        matches = self._get_matches(user_catch_all_pattern, ds_name)
        return next(matches, None)

    def match_runtime_pattern(self, ds_name: str) -> str:
        """
        Match a dataset name against the default runtime pattern.

        This method checks if the given dataset name matches any of the default
        runtime patterns. It assumes that a runtime pattern always matches.

        Args:
            ds_name: The name of the dataset to match.

        Returns:
            The first matching runtime pattern.

        Example:
        ``` python
        runtime_patterns = {"{default_example}": {"type": "MemoryDataset"}}
        resolver = CatalogConfigResolver(default_runtime_patterns=runtime_patterns)
        match = resolver.match_runtime_pattern("example_dataset")
        print(match)
        # {default_example}
        ```
        """
        default_patters = set(self._default_runtime_patterns.keys())
        matches = self._get_matches(default_patters, ds_name)
        # We assume runtime pattern always matches at the end
        return next(matches)

    def _get_pattern_config(self, pattern: str) -> dict[str, Any]:
        """
        Retrieve the configuration for a given dataset pattern.

        This method searches for the configuration of the specified pattern in the
        dataset patterns, user-set catch-all patterns, and runtime patterns. If the pattern is
        not found in any of these, an empty dictionary is returned.

        Args:
            pattern: The dataset pattern to retrieve the configuration for.

        Returns:
            The configuration dictionary for the specified pattern,
            or an empty dictionary if the pattern is not found.

        Example:
        ``` python
        resolver = CatalogConfigResolver(config={"example": {"type": "MemoryDataset"}})
        resolver._get_pattern_config("{default}")
        # {'type': 'kedro.io.MemoryDataset'}
        ```
        """
        return (
            self._dataset_patterns.get(pattern)
            or self._user_catch_all_pattern.get(pattern)
            or self._default_runtime_patterns.get(pattern)
            or {}
        )

    @classmethod
    def _extract_patterns(
        cls,
        config: dict[str, dict[str, Any]] | None,
        credentials: dict[str, dict[str, Any]] | None,
    ) -> tuple[Patterns, Patterns]:
        """
        Extract and sort patterns from the catalog configuration.

        This method identifies dataset patterns from the catalog configuration and
        sorts them based on specificity. It also identifies a catch-all pattern (if any)
        and sets it as the default pattern.

        Args:
            config: The catalog configuration containing dataset patterns.
            credentials: The credentials for datasets.

        Returns:
            A tuple containing:
                - Sorted dataset patterns.
                - The default pattern (if a catch-all pattern is found).

        Example:
        ``` python
        config = {
            "{namespace}.int_{name}": {"type": "pandas.CSVDataset"},
            "{name}": {"type": "MemoryDataset"},
        }
        sorted_patterns, default_pattern = CatalogConfigResolver._extract_patterns(
            config, None
        )
        print(sorted_patterns.keys())
        # dict_keys(['{namespace}.int_{name}'])
        print(default_pattern.keys())
        # dict_keys(['{name}'])
        ```
        """
        config = config or {}
        credentials = credentials or {}
        dataset_patterns = {}
        user_default = {}

        for ds_name, ds_config in config.items():
            if cls.is_pattern(ds_name):
                cls._validate_pattern_config(ds_name, ds_config)
                dataset_patterns[ds_name] = _resolve_credentials(ds_config, credentials)

        sorted_patterns = cls._sort_patterns(dataset_patterns)
        if sorted_patterns:
            # If the last pattern is a catch-all pattern, pop it and set it as the default
            if cls._pattern_specificity(list(sorted_patterns.keys())[-1]) == 0:
                last_pattern = sorted_patterns.popitem()
                user_default = {last_pattern[0]: last_pattern[1]}

        return sorted_patterns, user_default

    @classmethod
    def _resolve_credentials(
        cls,
        config: dict[str, dict[str, Any]] | None,
        credentials: dict[str, dict[str, Any]] | None,
    ) -> dict[str, dict[str, Any]]:
        """
        Resolve credentials for datasets in the catalog configuration.

        This method replaces credential references in the dataset configuration with
        the actual credentials from the provided credentials dictionary.

        Args:
            config: The catalog configuration containing datasets.
            credentials: The credentials for datasets.

        Returns:
            The dataset configurations with resolved credentials.

        Raises:
            DatasetError: If a dataset configuration is invalid (e.g., not a dictionary).

        Example:
        ``` python
        config = {
            "example": {
                "type": "pandas.CSVDataset",
                "credentials": "db_credentials",
            }
        }
        credentials = {"db_credentials": {"user": "username", "pass": "pass"}}
        resolved_configs = CatalogConfigResolver._resolve_credentials(
            config, credentials
        )
        print(resolved_configs)
        # {'example': {'type': 'pandas.CSVDataset', 'credentials': {'user': 'username', 'pass': 'pass'}}}
        ```
        """
        config = config or {}
        credentials = credentials or {}
        resolved_configs = {}

        for ds_name, ds_config in config.items():
            if not isinstance(ds_config, dict):
                raise DatasetError(
                    f"Catalog entry '{ds_name}' is not a valid dataset configuration. "
                    "\nHint: If this catalog entry is intended for variable interpolation, "
                    "make sure that the key is preceded by an underscore."
                )
            if not cls.is_pattern(ds_name):
                resolved_configs[ds_name] = _resolve_credentials(ds_config, credentials)

        return resolved_configs

    @staticmethod
    def _unresolve_credentials(
        cred_name: str, ds_config: dict[str, dict[str, Any]] | None
    ) -> tuple[dict[str, dict[str, Any]], dict[str, dict[str, Any]]]:
        """
        Extracts and replaces credentials in a dataset configuration with
        references, ensuring separation of credentials from the dataset configuration.

        Credentials are searched for recursively in the dataset configuration.
        The first occurrence of the `CREDENTIALS_KEY` is replaced with a generated
        reference key.

        Args:
            cred_name: A unique identifier for the credentials being unresolved.
                This is used to generate a reference key for the credentials.
            ds_config: The dataset configuration containing potential credentials
                under the key `CREDENTIALS_KEY`.

        Returns:
            A tuple containing:
                ds_config_copy : A deep copy of the original dataset
                    configuration with credentials replaced by reference keys.
                credentials: A dictionary mapping generated reference keys to the original credentials.

        Example:
        ``` python
        config = {
            "type": "pandas.CSVDataset",
            "credentials": {"user": "username", "pass": "pass"},
        }
        ds_config, creds = CatalogConfigResolver._unresolve_credentials(
            "example", config
        )
        print(ds_config)
        # {'type': 'pandas.CSVDataset', 'credentials': 'example_credentials'}
        print(creds)
        # {'example_credentials': {'user': 'username', 'pass': 'pass'}}
        ```
        """
        ds_config_copy = copy.deepcopy(ds_config) or {}
        credentials: dict[str, Any] = {}
        credentials_ref = f"{cred_name}_{CREDENTIALS_KEY}"

        def unresolve(config: Any) -> None:
            # We don't expect credentials key appears more than once within the same dataset config,
            # So once we found the key first time we unresolve it and stop iterating after
            for key, val in config.items():
                if key == CREDENTIALS_KEY and config[key]:
                    credentials[credentials_ref] = config[key]
                    config[key] = credentials_ref
                    return
                if isinstance(val, dict):
                    unresolve(val)

        unresolve(ds_config_copy)

        return ds_config_copy, credentials

    def resolve_pattern(self, ds_name: str) -> dict[str, Any]:
        """
        Resolve a dataset name to its configuration based on patterns.

        This method matches the dataset name against catalog patterns and resolves
        its configuration. If the dataset name matches a user-defined catch-all pattern, a warning
        is logged.

        Args:
            ds_name: The name of the dataset to resolve.

        Returns:
            The resolved dataset configuration.

        Example:
        ``` python
        config = {
            "{namespace}.int_{name}": {
                "type": "pandas.CSVDataset",
                "filepath": "{name}.csv",
            }
        }
        resolver = CatalogConfigResolver(config=config)
        resolved_config = resolver.resolve_pattern("data.int_customers")
        print(resolved_config)
        # {'type': 'pandas.CSVDataset', 'filepath': 'customers.csv'}
        ```
        """
        if ds_name not in self._resolved_configs:
            matched_pattern = (
                self.match_dataset_pattern(ds_name)
                or self.match_user_catch_all_pattern(ds_name)
                or self.match_runtime_pattern(ds_name)
            )
            pattern_config = self._get_pattern_config(matched_pattern)
            ds_config = self._resolve_dataset_config(
                ds_name, matched_pattern, copy.deepcopy(pattern_config)
            )

            if (
                self._pattern_specificity(matched_pattern) == 0
                and matched_pattern in self._user_catch_all_pattern
            ):
                self._logger.warning(
                    "Config from the dataset pattern '%s' in the catalog will be used to "
                    "override the default dataset creation for '%s'",
                    matched_pattern,
                    ds_name,
                )
            return ds_config  # type: ignore[no-any-return]

        return self._resolved_configs[ds_name]



"""``ParallelRunner`` is an ``AbstractRunner`` implementation. It can
be used to run the ``Pipeline`` in parallel groups formed by toposort.
"""

from __future__ import annotations

import os
from concurrent.futures import Executor, ProcessPoolExecutor
from multiprocessing import get_context
from multiprocessing.managers import SyncManager
from multiprocessing.reduction import ForkingPickler
from pickle import PicklingError
from typing import TYPE_CHECKING

from kedro.io import (
    MemoryDataset,
)
from kedro.runner.runner import AbstractRunner

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pluggy import PluginManager

    from kedro.io import SharedMemoryCatalogProtocol
    from kedro.pipeline import Pipeline
    from kedro.pipeline.node import Node


class ParallelRunnerManager(SyncManager):
    """``ParallelRunnerManager`` is used to create shared ``MemoryDataset``
    objects as default datasets in a pipeline.
    """


ParallelRunnerManager.register("MemoryDataset", MemoryDataset)


class ParallelRunner(AbstractRunner):
    """``ParallelRunner`` is an ``AbstractRunner`` implementation. It can
    be used to run the ``Pipeline`` in parallel groups formed by toposort.
    Please note that this `runner` implementation validates dataset using the
    ``_validate_catalog`` method, which checks if any of the datasets are
    single process only using the `_SINGLE_PROCESS` dataset attribute.
    """

    def __init__(
        self,
        max_workers: int | None = None,
        is_async: bool = False,
    ):
        """
        Instantiates the runner by creating a Manager.

        Args:
            max_workers: Number of worker processes to spawn. If not set,
                calculated automatically based on the pipeline configuration
                and CPU core count. On windows machines, the max_workers value
                cannot be larger than 61 and will be set to min(61, max_workers).
            is_async: If True, the node inputs and outputs are loaded and saved
                asynchronously with threads. Defaults to False.
        Raises:
            ValueError: bad parameters passed
        """
        super().__init__(is_async=is_async)
        self._manager = ParallelRunnerManager()
        self._manager.start()

        self._max_workers = self._validate_max_workers(max_workers)

    def __del__(self) -> None:
        self._manager.shutdown()

    @classmethod
    def _validate_nodes(cls, nodes: Iterable[Node]) -> None:
        """Ensure all tasks are serialisable."""
        unserialisable = []
        for node in nodes:
            try:
                ForkingPickler.dumps(node)
            except (AttributeError, PicklingError):
                unserialisable.append(node)

        if unserialisable:
            raise AttributeError(
                f"The following nodes cannot be serialised: {sorted(unserialisable)}\n"
                f"In order to utilize multiprocessing you need to make sure all nodes "
                f"are serialisable, i.e. nodes should not include lambda "
                f"functions, nested functions, closures, etc.\nIf you "
                f"are using custom decorators ensure they are correctly decorated using "
                f"functools.wraps()."
            )

    @classmethod
    def _validate_catalog(cls, catalog: SharedMemoryCatalogProtocol) -> None:  # type: ignore[override]
        """Ensure that all datasets are serialisable and that we do not have
        any non proxied memory datasets being used as outputs as their content
        will not be synchronized across threads.
        """
        catalog.validate_catalog()

    def _set_manager_datasets(self, catalog: SharedMemoryCatalogProtocol) -> None:  # type: ignore[override]
        catalog.set_manager_datasets(self._manager)

    def _get_required_workers_count(self, pipeline: Pipeline) -> int:
        """
        Calculate the max number of processes required for the pipeline,
        limit to the number of CPU cores.
        """
        # Number of nodes is a safe upper-bound estimate.
        # It's also safe to reduce it by the number of layers minus one,
        # because each layer means some nodes depend on other nodes
        # and they can not run in parallel.
        # It might be not a perfect solution, but good enough and simple.
        required_processes = len(pipeline.nodes) - len(pipeline.grouped_nodes) + 1

        return min(required_processes, self._max_workers)

    def _get_executor(self, max_workers: int) -> Executor:
        context = os.environ.get("KEDRO_MP_CONTEXT")
        if context and context not in {"fork", "spawn"}:
            context = None
        ctx = get_context(context)
        return ProcessPoolExecutor(max_workers=max_workers, mp_context=ctx)

    def _run(
        self,
        pipeline: Pipeline,
        catalog: SharedMemoryCatalogProtocol,  # type: ignore[override]
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
    ) -> None:
        """The method implementing parallel pipeline running.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: An implemented instance of ``SharedMemoryCatalogProtocol`` from which to fetch data.
            hook_manager: The ``PluginManager`` to activate hooks.
            run_id: The id of the run.

        Raises:
            AttributeError: When the provided pipeline is not suitable for
                parallel execution.
            RuntimeError: If the runner is unable to schedule the execution of
                all pipeline nodes.
            Exception: In case of any downstream node failure.

        """
        if not self._is_async:
            self._logger.info(
                "Using synchronous mode for loading and saving data. Use the --async flag "
                "for potential performance gains. https://docs.kedro.org/en/stable/nodes_and_pipelines/run_a_pipeline.html#load-and-save-asynchronously"
            )

        super()._run(
            pipeline=pipeline,
            catalog=catalog,
            run_id=run_id,
        )



import tempfile
from pathlib import Path

from kedro.config import OmegaConfigLoader


# Helper functions to generate sample configuration data
def _generate_catalog(start_range, end_range, is_local=False, is_versioned=False, add_interpolation=False):
    catalog = {}
    for i in range(start_range, end_range + 1):
        catalog[f"dataset_{i}"] = {
            "type": "pandas.CSVDataset",
            "filepath": f"data{i}{'_local' if is_local else ''}.csv"
        }
        if is_versioned:
            catalog[f"dataset_{i}"]["versioned"] = True
        if add_interpolation:
            catalog[f"dataset_{i}"]["filepath"] = "${_basepath}" + catalog[f"dataset_{i}"]["filepath"]
    return catalog

def _generate_params(start_range, end_range, is_local=False, add_globals=False):
    if add_globals:
        # Generate params with "${globals:global{i}}"
        params = {f"param_{i}": f"${{globals:global_{i}}}" for i in range(start_range, end_range + 1)}
    else:
        # Generate params with "value_{i}" or "value_{i}_local"
        params = {f"param_{i}": f"value_{i}{'_local' if is_local else ''}" for i in range(start_range, end_range + 1)}

    return params

def _generate_globals(start_range, end_range, is_local=False):
    globals_dict = {f"global_{i}": f"value{i}{'_local' if is_local else ''}" for i in range(start_range, end_range + 1)}
    return globals_dict

def _create_config_file(conf_source, env, file_name, data):
    env_path = conf_source / env
    env_path.mkdir(parents=True, exist_ok=True)
    file_path = env_path / file_name

    import yaml
    with open(file_path, "w") as f:
        yaml.dump(data, f)

base_catalog = _generate_catalog(1, 1000, is_versioned=True)
local_catalog = _generate_catalog(501, 1500, is_local=True)
base_params = _generate_params(1, 1000)
local_params = _generate_params(501, 1500, is_local=True)
base_globals = _generate_globals(1, 1000)
local_globals = _generate_globals(501, 1500, is_local=True)

base_catalog_with_interpolations = _generate_catalog(1, 1000, is_versioned=True, add_interpolation=True)
base_catalog_with_interpolations.update({"_basepath": "/path/to/data"})
local_catalog_with_interpolations = _generate_catalog(501, 1500, is_local=True, add_interpolation=True)
local_catalog_with_interpolations.update({"_basepath": "/path/to/data"})

base_params_with_globals = _generate_params(1, 100, add_globals=True)


class TimeOmegaConfigLoader:
    def setup(self):
        # Setup temporary configuration directory with sample config files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.conf_source = Path(self.temp_dir.name)

        # Create sample config files in the temp directory
        _create_config_file(self.conf_source, "base", "catalog.yml", base_catalog)
        _create_config_file(self.conf_source, "local", "catalog.yml", local_catalog)
        _create_config_file(self.conf_source, "base", "parameters.yml", base_params)
        _create_config_file(self.conf_source, "local", "parameters.yml", local_params)
        _create_config_file(self.conf_source, "base", "globals.yml", base_globals)
        _create_config_file(self.conf_source, "local", "globals.yml", local_globals)

        # Instantiate the OmegaConfigLoader
        self.loader = OmegaConfigLoader(conf_source=self.conf_source, base_env='base', default_run_env='local')

    def teardown(self):
        # Cleanup temporary directory
        self.temp_dir.cleanup()

    def time_loading_catalog(self):
        """Benchmark the time to load the catalog"""
        self.loader["catalog"]

    def time_loading_parameters(self):
        """Benchmark the time to load the parameters"""
        self.loader["parameters"]

    def time_loading_globals(self):
        """Benchmark the time to load global configuration"""
        self.loader["globals"]

    def time_loading_parameters_runtime(self):
        """Benchmark the time to load parameters with runtime configuration"""
        self.loader.runtime_params = _generate_params(2001, 2002)
        self.loader["parameters"]

    def time_merge_soft_strategy(self):
        """Benchmark the time to load and soft-merge configurations"""
        self.loader.merge_strategy = {"catalog": "soft"}
        self.loader["catalog"]


class TimeOmegaConfigLoaderAdvanced:
    def setup(self):
        # Setup temporary configuration directory with sample config files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.conf_source = Path(self.temp_dir.name)

        # Create sample config files in the temp directory
        _create_config_file(self.conf_source, "base", "catalog.yml", base_catalog_with_interpolations)
        _create_config_file(self.conf_source, "local", "catalog.yml", local_catalog_with_interpolations)
        _create_config_file(self.conf_source, "base", "parameters.yml", base_params_with_globals)
        _create_config_file(self.conf_source, "base", "globals.yml", base_globals)

        # Instantiate the OmegaConfigLoader
        self.loader = OmegaConfigLoader(conf_source=self.conf_source, base_env='base', default_run_env='local')

    def teardown(self):
        # Cleanup temporary directory
        self.temp_dir.cleanup()

    def time_loading_catalog(self):
        """Benchmark the time to load the catalog"""
        self.loader["catalog"]

    def time_loading_parameters(self):
        """Benchmark the time to load parameters with global interpolation"""
        self.loader["parameters"]



"""``ThreadRunner`` is an ``AbstractRunner`` implementation. It can
be used to run the ``Pipeline`` in parallel groups formed by toposort
using threads.
"""

from __future__ import annotations

import warnings
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import TYPE_CHECKING

from kedro.runner.runner import AbstractRunner

if TYPE_CHECKING:
    from pluggy import PluginManager

    from kedro.io import CatalogProtocol
    from kedro.pipeline import Pipeline


class ThreadRunner(AbstractRunner):
    """``ThreadRunner`` is an ``AbstractRunner`` implementation. It can
    be used to run the ``Pipeline`` in parallel groups formed by toposort
    using threads.
    """

    def __init__(
        self,
        max_workers: int | None = None,
        is_async: bool = False,
    ):
        """
        Instantiates the runner.

        Args:
            max_workers: Number of worker processes to spawn. If not set,
                calculated automatically based on the pipeline configuration
                and CPU core count.
            is_async: If True, set to False, because `ThreadRunner`
                doesn't support loading and saving the node inputs and
                outputs asynchronously with threads. Defaults to False.
        Raises:
            ValueError: bad parameters passed
        """
        if is_async:
            warnings.warn(
                "'ThreadRunner' doesn't support loading and saving the "
                "node inputs and outputs asynchronously with threads. "
                "Setting 'is_async' to False."
            )
        super().__init__(is_async=False)

        self._max_workers = self._validate_max_workers(max_workers)

    def _get_required_workers_count(self, pipeline: Pipeline) -> int:
        """
        Calculate the max number of processes required for the pipeline
        """
        # Number of nodes is a safe upper-bound estimate.
        # It's also safe to reduce it by the number of layers minus one,
        # because each layer means some nodes depend on other nodes
        # and they can not run in parallel.
        # It might be not a perfect solution, but good enough and simple.
        required_threads = len(pipeline.nodes) - len(pipeline.grouped_nodes) + 1

        return (
            min(required_threads, self._max_workers)
            if self._max_workers
            else required_threads
        )

    def _get_executor(self, max_workers: int) -> Executor:
        return ThreadPoolExecutor(max_workers=max_workers)

    def _run(
        self,
        pipeline: Pipeline,
        catalog: CatalogProtocol,
        hook_manager: PluginManager | None = None,
        run_id: str | None = None,
    ) -> None:
        """The method implementing threaded pipeline running.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: An implemented instance of ``CatalogProtocol`` from which to fetch data.
            hook_manager: The ``PluginManager`` to activate hooks.
            run_id: The id of the run.

        Raises:
            Exception: in case of any downstream node failure.

        """
        super()._run(
            pipeline=pipeline,
            catalog=catalog,
            hook_manager=hook_manager,
            run_id=run_id,
        )



"""``kedro.runner`` provides runners that are able
to execute ``Pipeline`` instances.
"""

from .parallel_runner import ParallelRunner
from .runner import AbstractRunner
from .sequential_runner import SequentialRunner
from .task import Task
from .thread_runner import ThreadRunner

__all__ = [
    "AbstractRunner",
    "ParallelRunner",
    "SequentialRunner",
    "Task",
    "ThreadRunner",
]



"""
This module contains example tests for a Kedro project.
Tests should be placed in ``src/tests``, in modules that mirror your
project's structure, and in files named test_*.py.
"""
import pytest
from pathlib import Path
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project

# The tests below are here for the demonstration purpose
# and should be replaced with the ones testing the project
# functionality

class TestKedroRun:
    def test_kedro_run_no_pipeline(self):
    # This example test expects a pipeline run failure, since
    # the default project template contains no pipelines.
        bootstrap_project(Path.cwd())

        with pytest.raises(Exception) as excinfo:
            with KedroSession.create(project_path=Path.cwd()) as session:
                session.run()

        assert "Pipeline contains no nodes" in str(excinfo.value)



import logging
from collections.abc import Iterable
from typing import Any

from kedro.framework.project import pipelines as _pipelines
from kedro.io import DataCatalog
from kedro.io.core import is_parameter
from kedro.pipeline import Pipeline


class CatalogCommandsMixin:
    """
    A mixin class that provides additional commands for interacting with the `DataCatalog`.

    This class adds methods to list datasets, dataset factory patterns and resolve dataset factory patterns.
    It is designed to extend the functionality of the `DataCatalog` providing pipeline-based
    catalog functionality.

    Methods:
        - describe_datasets: Show datasets per type for specified pipelines.
        - list_patterns: List all dataset factory patterns in the catalog.
        - resolve_patterns: Resolve dataset factories against pipeline datasets.

    Usage:

    You can integrate this mixin with the `DataCatalog` in two ways:

    1. Using `compose_classes`:
    ``` python
    from kedro.io import DataCatalog
    from kedro.framework.context import CatalogCommandsMixin, compose_classes

    # DataCatalog instance without CatalogCommandsMixin
    assert not hasattr(DataCatalog(), "describe_datasets")

    # Compose a new class combining DataCatalog and CatalogCommandsMixin
    catalog_class = compose_classes(DataCatalog, CatalogCommandsMixin)

    # Create a catalog instance from configuration
    catalog = catalog_class.from_config(
        {
            "cars": {
                "type": "pandas.CSVDataset",
                "filepath": "cars.csv",
                "save_args": {"index": False},
            }
        }
    )

    # Assert that the catalog has the `describe_datasets` method
    assert hasattr(
        catalog, "describe_datasets"
    ), "describe_datasets method is not available"
    print("describe_datasets method is available!")
    # describe_datasets method is available!
    ```

    2. Creating a new class with inheritance:
    ``` python
    from kedro.io import DataCatalog
    from kedro.framework.context import CatalogCommandsMixin


    class DataCatalogWithMixins(DataCatalog, CatalogCommandsMixin):
        pass


    catalog = DataCatalogWithMixins(datasets={"example": MemoryDataset()})
    assert hasattr(
        catalog, "describe_datasets"
    ), "describe_datasets method is not available"
    print("describe_datasets method is available!")
    # describe_datasets method is available!
    ```
    """

    @property
    def _logger(self) -> logging.Logger: ...  # type: ignore[empty-body]

    def describe_datasets(
        self: DataCatalog, pipelines: list[str] | list[Pipeline] | None = None
    ) -> dict:
        """
        Describe datasets used in the specified pipelines, grouped by type.

        This method provides a structured summary of datasets used in the selected pipelines,
        categorizing them into three groups:
        - `datasets`: Datasets explicitly defined in the catalog.
        - `factories`: Datasets resolved from dataset factory patterns.
        - `defaults`: Datasets that do not match any pattern or explicit definition.

        Args:
            pipelines: A list of pipeline names or `Pipeline` objects to analyze.
                If `None`, all pipelines are analyzed.

        Returns:
            A dictionary where keys are pipeline names and values are dictionaries
            containing datasets grouped by type.

        Example output:
        {
            "data_processing": {
                "datasets": {
                    "kedro_datasets.pandas.parquet_dataset.ParquetDataset": ["model_input_table"]
                },
                "factories": {},
                "defaults": {"kedro.io.MemoryDataset": ["preprocessed_companies"]}
            }
        }
        """
        target_pipelines = pipelines or _pipelines.keys()

        result = {}
        if not isinstance(target_pipelines, Iterable):
            target_pipelines = [target_pipelines]

        for i, pipe in enumerate(target_pipelines):
            pipeline_ds = set()
            pl_obj = _pipelines.get(pipe) if isinstance(pipe, str) else pipe
            if pl_obj:
                pipeline_ds = pl_obj.datasets()

            catalog_ds = set(self.keys())

            patterns_ds = set()
            default_ds = set()
            for ds_name in pipeline_ds - catalog_ds:
                if self.config_resolver.match_dataset_pattern(ds_name):
                    patterns_ds.add(ds_name)
                else:
                    default_ds.add(ds_name)

            used_ds_by_type = _group_ds_by_type(
                pipeline_ds - patterns_ds - default_ds, self
            )
            patterns_ds_by_type = _group_ds_by_type(patterns_ds, self)
            default_ds_by_type = _group_ds_by_type(default_ds, self)

            data = (
                ("datasets", used_ds_by_type),
                ("factories", patterns_ds_by_type),
                ("defaults", default_ds_by_type),
            )
            pipe_name = pipe if isinstance(pipe, str) else f"pipeline_{i}"
            result[pipe_name] = {key: value for key, value in data}

        return result

    def list_patterns(self: DataCatalog) -> list[str]:
        """
        List all dataset factory patterns in the catalog, ranked by priority.

        This method retrieves all dataset factory patterns defined in the catalog,
        ordered by the priority in which they are matched.

        Returns:
            A list of dataset factory patterns.
        """
        return self.config_resolver.list_patterns()

    def resolve_patterns(
        self: DataCatalog,
        pipelines: list[Pipeline] | None = None,
    ) -> dict[str, Any]:
        """
        Resolve dataset factory patterns against pipeline datasets.

        This method resolves dataset factory patterns for datasets used in the specified pipelines.
        It includes datasets explicitly defined in the catalog as well as those resolved
        from dataset factory patterns.

        Args:
            pipelines: A list of `Pipeline` objects to analyze.
                If `None`, all pipelines are analyzed.

        Returns:
            A dictionary mapping dataset names to their unresolved configurations.
        """
        target_pipelines = pipelines or _pipelines.keys()

        pipeline_datasets: set[str] = set()
        for pipe in target_pipelines:
            pl_obj = _pipelines.get(pipe) if isinstance(pipe, str) else pipe
            if pl_obj:
                pipeline_datasets.update(pl_obj.datasets())

        # We need to include datasets defined in the catalog.yaml and datasets added manually to the catalog
        explicit_datasets = {}
        for ds_name, ds in self.items():
            if is_parameter(ds_name):
                continue

            unresolved_config, _ = self.config_resolver._unresolve_credentials(
                ds_name, ds._init_config()
            )
            explicit_datasets[ds_name] = unresolved_config

        for ds_name in pipeline_datasets:
            if ds_name in explicit_datasets or is_parameter(ds_name):
                continue

            ds_config = self.config_resolver.resolve_pattern(ds_name)
            unresolved_config, _ = self.config_resolver._unresolve_credentials(
                ds_name, ds_config
            )
            explicit_datasets[ds_name] = unresolved_config

        return explicit_datasets


def _group_ds_by_type(datasets: set[str], catalog: DataCatalog) -> dict[str, list[str]]:
    """
    Group datasets by their type.

    This utility function groups datasets into a dictionary where the keys are
    dataset types (as strings) and the values are lists of dataset names.

    Args:
        datasets: A set of dataset names to group.
        catalog: The `DataCatalog` instance to retrieve dataset types.

    Returns:
        A dictionary mapping dataset types to lists of dataset names.

    Example:
    ``` python
    from kedro.io.data_catalog import DataCatalog
    from kedro.io.memory_dataset import MemoryDataset
    from kedro.framework.context.catalog_mixins import _group_ds_by_type

    catalog = DataCatalog(datasets={"example": MemoryDataset()})
    datasets = {"example"}
    _group_ds_by_type(datasets, catalog)
    # {'kedro.io.memory_dataset.MemoryDataset': ['example']}
    ```
    """
    mapping: dict[str, list[str]] = {}
    for ds_name in datasets:
        if is_parameter(ds_name):
            continue

        str_type = None
        if ds_name in catalog:
            str_type = catalog.get_type(ds_name)

        if str_type is None:
            str_type = catalog.default_runtime_patterns["{default}"]["type"]

        if str_type not in mapping:
            mapping[str_type] = []

        mapping[str_type].append(ds_name)

    return mapping



"""
This script creates an IPython extension to load Kedro-related variables in
local scope.
"""

from __future__ import annotations

import importlib
import inspect
import logging
import os
import sys
import typing
import warnings
from pathlib import Path
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Final

if TYPE_CHECKING:
    from collections import OrderedDict
    from collections.abc import Callable

    from IPython.core.interactiveshell import InteractiveShell

from IPython.core.getipython import get_ipython
from IPython.core.magic import needs_local_scope, register_line_magic
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring

try:
    import rich.console as rich_console
    import rich.syntax as rich_syntax
except ImportError:  # pragma: no cover
    pass

from kedro.framework.cli import load_entry_points
from kedro.framework.cli.project import CONF_SOURCE_HELP, PARAMS_ARG_HELP
from kedro.framework.cli.utils import ENV_HELP, _split_params
from kedro.framework.project import (
    LOGGING,  # noqa: F401
    _ProjectPipelines,
    configure_project,
    pipelines,
)
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from kedro.pipeline.node import Node
from kedro.utils import _is_databricks, find_kedro_project

logger = logging.getLogger(__name__)

FunctionParameters = MappingProxyType

RICH_INSTALLED: Final = importlib.util.find_spec("rich") is not None


def load_ipython_extension(ipython: InteractiveShell) -> None:
    """
    Main entry point when %load_ext kedro.ipython is executed, either manually or
    automatically through `kedro ipython` or `kedro jupyter lab/notebook`.
    IPython will look for this function specifically.
    See https://ipython.readthedocs.io/en/stable/config/extensions/index.html
    """
    ipython.register_magic_function(func=magic_reload_kedro, magic_name="reload_kedro")  # type: ignore[call-arg]
    logger.info("Registered line magic '%reload_kedro'")
    ipython.register_magic_function(func=magic_load_node, magic_name="load_node")  # type: ignore[call-arg]
    logger.info("Registered line magic '%load_node'")

    if find_kedro_project(Path.cwd()) is None:
        logger.warning(
            "Kedro extension was registered but couldn't find a Kedro project. "
            "Make sure you run '%reload_kedro <project_root>'."
        )
        return

    reload_kedro()


@typing.no_type_check
@needs_local_scope
@magic_arguments()
@argument(
    "path",
    type=str,
    help=(
        "Path to the project root directory. If not given, use the previously set"
        "project root."
    ),
    nargs="?",
    default=None,
)
@argument("-e", "--env", type=str, default=None, help=ENV_HELP)
@argument(
    "--params",
    type=lambda value: _split_params(None, None, value),
    default=None,
    help=PARAMS_ARG_HELP,
)
@argument("--conf-source", type=str, default=None, help=CONF_SOURCE_HELP)
def magic_reload_kedro(
    line: str,
    local_ns: dict[str, Any] | None = None,
    conf_source: str | None = None,
) -> None:
    """
    The `%reload_kedro` IPython line magic.
    See https://docs.kedro.org/en/stable/notebooks_and_ipython/kedro_and_notebooks.html#reload-kedro-line-magic
    for more.
    """
    args = parse_argstring(magic_reload_kedro, line)
    reload_kedro(args.path, args.env, args.params, local_ns, args.conf_source)


def reload_kedro(
    path: str | None = None,
    env: str | None = None,
    runtime_params: dict[str, Any] | None = None,
    local_namespace: dict[str, Any] | None = None,
    conf_source: str | None = None,
) -> None:  # pragma: no cover
    """Function that underlies the %reload_kedro Line magic. This should not be imported
    or run directly but instead invoked through %reload_kedro."""

    project_path = _resolve_project_path(path, local_namespace)

    metadata = bootstrap_project(project_path)
    _remove_cached_modules(metadata.package_name)
    configure_project(metadata.package_name)

    session = KedroSession.create(
        project_path,
        env=env,
        runtime_params=runtime_params,
        conf_source=conf_source,
    )
    context = session.load_context()
    catalog = context.catalog

    get_ipython().push(  # type: ignore[no-untyped-call]
        variables={
            "context": context,
            "catalog": catalog,
            "session": session,
            "pipelines": pipelines,
        }
    )

    logger.info("Kedro project %s", str(metadata.project_name))
    logger.info(
        "Defined global variable 'context', 'session', 'catalog' and 'pipelines'"
    )

    for line_magic in load_entry_points("line_magic"):
        register_line_magic(needs_local_scope(line_magic))  # type: ignore[no-untyped-call]
        logger.info("Registered line magic '%s'", line_magic.__name__)  # type: ignore[attr-defined]


def _resolve_project_path(
    path: str | None = None, local_namespace: dict[str, Any] | None = None
) -> Path:
    """
    Resolve the project path to use with reload_kedro, updating or adding it
    (in-place) to the local ipython Namespace (``local_namespace``) if necessary.

    Arguments:
        path: the path to use as a string object
        local_namespace: Namespace with local variables of the scope where the line
            magic is invoked in a dict.
    """
    if path:
        project_path = Path(path).expanduser().resolve()
    else:
        if (
            local_namespace
            and local_namespace.get("context")
            and hasattr(local_namespace["context"], "project_path")
        ):
            project_path = local_namespace["context"].project_path
        else:
            project_path = find_kedro_project(Path.cwd())
        if project_path:
            logger.info(
                "Resolved project path as: %s.\nTo set a different path, run "
                "'%%reload_kedro <project_root>'",
                project_path,
            )

    if (
        project_path
        and local_namespace
        and local_namespace.get("context")
        and hasattr(local_namespace["context"], "project_path")  # Avoid name collision
        and project_path != local_namespace["context"].project_path
    ):
        logger.info("Updating path to Kedro project: %s...", project_path)

    return project_path


def _remove_cached_modules(package_name: str) -> None:  # pragma: no cover
    to_remove = [mod for mod in sys.modules if mod.startswith(package_name)]
    # `del` is used instead of `reload()` because: If the new version of a module does not
    # define a name that was defined by the old version, the old definition remains.
    for module in to_remove:
        del sys.modules[module]


def _guess_run_environment() -> str:  # pragma: no cover
    """Best effort to guess the IPython/Jupyter environment"""
    # https://github.com/microsoft/vscode-jupyter/issues/7380
    if os.environ.get("VSCODE_PID") or os.environ.get("VSCODE_CWD"):
        return "vscode"
    elif _is_databricks():
        return "databricks"
    elif hasattr(get_ipython(), "kernel"):  # type: ignore[no-untyped-call]
        # IPython terminal does not have this attribute
        return "jupyter"
    else:
        return "ipython"


@typing.no_type_check
@magic_arguments()
@argument(
    "node",
    type=str,
    help=("Name of the Node."),
    nargs="?",
    default=None,
)
def magic_load_node(args: str) -> None:
    """The line magic %load_node <node_name>.
    Currently, this feature is only available for Jupyter Notebook (>7.0), Jupyter Lab, IPython,
    and VSCode Notebook. This line magic will generate code in multiple cells to load
    datasets from `DataCatalog`, import relevant functions and modules, node function
    definition and a function call. If generating code is not possible, it will print
    the code instead.
    """
    parameters = parse_argstring(magic_load_node, args)
    node_name = parameters.node

    cells = _load_node(node_name, pipelines)

    run_environment = _guess_run_environment()

    if run_environment in ("ipython", "vscode", "jupyter"):
        # Combine multiple cells into one for IPython or VSCode or Jupyter
        combined_cell = "\n\n".join(cells)
        _create_cell_with_text(combined_cell)
    else:
        # For other environments or if detection fails, just print the cells
        _print_cells(cells)


class _NodeBoundArguments(inspect.BoundArguments):
    """Similar to inspect.BoundArguments"""

    def __init__(
        self, signature: inspect.Signature, arguments: OrderedDict[str, Any]
    ) -> None:
        super().__init__(signature, arguments)

    @property
    def input_params_dict(self) -> dict[str, str] | None:
        """A mapping of {variable name: dataset_name}"""
        var_positional_arg_name = self._find_var_positional_arg()
        inputs_params_dict = {}
        for param, dataset_name in self.arguments.items():
            if param == var_positional_arg_name:
                # If the argument is *args, use the dataset name instead
                for arg in dataset_name:
                    inputs_params_dict[arg] = arg
            else:
                inputs_params_dict[param] = dataset_name
        return inputs_params_dict

    def _find_var_positional_arg(self) -> str | None:
        """Find the name of the VAR_POSITIONAL argument( *args), if any."""
        for k, v in self.signature.parameters.items():
            if v.kind == inspect.Parameter.VAR_POSITIONAL:
                return k
        return None


def _create_cell_with_text(text: str) -> None:
    """Create a new cell with the provided text content."""
    get_ipython().set_next_input(text)  # type: ignore[no-untyped-call]


def _print_cells(cells: list[str]) -> None:
    for cell in cells:
        if RICH_INSTALLED:
            rich_console.Console().print("")
            rich_console.Console().print(
                rich_syntax.Syntax(cell, "python", theme="monokai", line_numbers=False)
            )
        else:
            print("")  # noqa: T201
            print(cell)  # noqa: T201


def _load_node(node_name: str, pipelines: _ProjectPipelines) -> list[str]:
    """Prepare the code to load dataset from catalog, import statements and function body.

    Args:
        node_name (str): The name of the node.

    Returns:
        list[str]: A list of string which is the generated code, each string represent a
        notebook cell.
    """
    warnings.warn(
        "This is an experimental feature, only Jupyter Notebook (>7.0), Jupyter Lab, IPython, and VSCode Notebook "
        "are supported. If you encounter unexpected behaviour or would like to suggest "
        "feature enhancements, add it under this github issue https://github.com/kedro-org/kedro/issues/3580"
    )
    node = _find_node(node_name, pipelines)
    node_func = node.func

    imports_cell = _prepare_imports(node_func)
    function_definition_cell = _prepare_function_body(node_func)

    node_bound_arguments = _get_node_bound_arguments(node)
    inputs_params_mapping = _prepare_node_inputs(node_bound_arguments)
    node_inputs_cell = _format_node_inputs_text(inputs_params_mapping)
    function_call_cell = _prepare_function_call(node_func, node_bound_arguments)

    cells: list[str] = []
    if node_inputs_cell:
        cells.append(node_inputs_cell)
    cells.append(imports_cell)
    cells.append(function_definition_cell)
    cells.append(function_call_cell)
    return cells


def _find_node(node_name: str, pipelines: _ProjectPipelines) -> Node:
    for pipeline in pipelines.values():
        try:
            found_node: Node = pipeline.filter(node_names=[node_name]).nodes[0]
            return found_node
        except ValueError:
            continue
    # If reached the node was not found in the project
    raise ValueError(
        f"Node with name='{node_name}' not found in any pipelines. Remember to specify the node name, not the node function."
    )


def _prepare_imports(node_func: Callable) -> str:
    """Prepare the import statements for loading a node."""
    python_file = inspect.getsourcefile(node_func)
    logger.info(f"Loading node definition from {python_file}")

    # Confirm source file was found
    if python_file:
        import_statement = []
        with open(python_file) as file:
            # Handle multiline imports, i.e.
            # from lib import (
            # a,
            # b,
            # c
            # )
            # This will not work with all edge cases but good enough with common cases that
            # are formatted automatically by black, ruff etc.
            inside_bracket = False
            # Parse any line start with from or import statement

            for _ in file.readlines():
                line = _.strip()
                if not inside_bracket:
                    # The common case
                    if line.startswith("from") or line.startswith("import"):
                        import_statement.append(line)
                        if line.endswith("("):
                            inside_bracket = True
                # Inside multi-lines import, append everything.
                else:
                    import_statement.append(line)
                    if line.endswith(")"):
                        inside_bracket = False

        clean_imports = "\n".join(import_statement).strip()
        return clean_imports
    else:
        raise FileNotFoundError(f"Could not find {node_func.__name__}")


def _get_node_bound_arguments(node: Node) -> _NodeBoundArguments:
    node_func = node.func
    node_inputs = node.inputs

    args, kwargs = Node._process_inputs_for_bind(node_inputs)
    signature = inspect.signature(node_func)
    bound_arguments = signature.bind(*args, **kwargs)
    return _NodeBoundArguments(bound_arguments.signature, bound_arguments.arguments)


def _prepare_node_inputs(
    node_bound_arguments: _NodeBoundArguments,
) -> dict[str, str] | None:
    # Remove the *args. For example {'first_arg':'a', 'args': ('b','c')}
    # will be loaded as follow:
    # first_arg = catalog.load("a")
    # b = catalog.load("b") # It doesn't have an arg name, so use the dataset name instead.
    # c = catalog.load("c")
    return node_bound_arguments.input_params_dict


def _format_node_inputs_text(input_params_dict: dict[str, str] | None) -> str | None:
    statements = [
        "# Prepare necessary inputs for debugging",
        "# All debugging inputs must be defined in your project catalog",
    ]
    if not input_params_dict:
        return None

    for func_param, dataset_name in input_params_dict.items():
        statements.append(f'{func_param} = catalog.load("{dataset_name}")')

    input_statements = "\n".join(statements)
    return input_statements


def _prepare_function_body(func: Callable) -> str:
    source_lines, _ = inspect.getsourcelines(func)
    body = "".join(source_lines)
    return body


def _prepare_function_call(
    node_func: Callable, node_bound_arguments: _NodeBoundArguments
) -> str:
    """Prepare the text for the function call."""
    func_name = node_func.__name__
    args = node_bound_arguments.input_params_dict
    kwargs = node_bound_arguments.kwargs

    # Construct the statement of func_name(a=1,b=2,c=3)
    args_str_literal = [f"{node_input}" for node_input in args] if args else []
    kwargs_str_literal = [
        f"{node_input}={dataset_name}" for node_input, dataset_name in kwargs.items()
    ]
    func_params = ", ".join(args_str_literal + kwargs_str_literal)
    body = f"""{func_name}({func_params})"""
    return body



"""This module provides context for Kedro project."""

from __future__ import annotations

import logging
from copy import deepcopy
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse
from warnings import warn

from attrs import define, field

from kedro.config import AbstractConfigLoader, MissingConfigException
from kedro.framework.context import CatalogCommandsMixin
from kedro.io import CatalogProtocol, DataCatalog
from kedro.pipeline.transcoding import _transcode_split

if TYPE_CHECKING:
    from pluggy import PluginManager


def _is_relative_path(path_string: str) -> bool:
    """Checks whether a path string is a relative path.

    Example:
    ``` python
    _is_relative_path("data/01_raw") == True
    _is_relative_path("info.log") == True
    _is_relative_path("/tmp/data/01_raw") == False
    _is_relative_path(r"C:\\info.log") == False
    _is_relative_path(r"\\'info.log") == False
    _is_relative_path("c:/info.log") == False
    _is_relative_path("s3://info.log") == False

    Args:
        path_string: The path string to check.
    Returns:
        Whether the string is a relative path.
    ```
    """
    # os.path.splitdrive does not reliably work on non-Windows systems
    # breaking the coverage, using PureWindowsPath instead
    is_full_windows_path_with_drive = bool(PureWindowsPath(path_string).drive)
    if is_full_windows_path_with_drive:
        return False

    is_remote_path = bool(urlparse(path_string).scheme)
    if is_remote_path:
        return False

    is_absolute_path = PurePosixPath(path_string).is_absolute()
    if is_absolute_path:
        return False

    return True


def _convert_paths_to_absolute_posix(
    project_path: Path, conf_dictionary: dict[str, Any]
) -> dict[str, Any]:
    """Turn all relative paths inside ``conf_dictionary`` into absolute paths by appending them
    to ``project_path`` and convert absolute Windows paths to POSIX format. This is a hack to
    make sure that we don't have to change user's working directory for logging and datasets to
    work. It is important for non-standard workflows such as IPython notebook where users don't go
    through `kedro run` or `__main__.py` entrypoints.

    Example:
    ```` python
    conf = _convert_paths_to_absolute_posix(
        project_path=Path("/path/to/my/project"),
        conf_dictionary={
            "handlers": {
                "info_file_handler": {
                    "filename": "info.log"
                }
            }
        }
    )
    print(conf['handlers']['info_file_handler']['filename'])
    "/path/to/my/project/info.log"

    Args:
        project_path: The root directory to prepend to relative path to make absolute path.
        conf_dictionary: The configuration containing paths to expand.
    Returns:
        A dictionary containing only absolute paths.
    Raises:
        ValueError: If the provided ``project_path`` is not an absolute path.
    ```
    """
    if not project_path.is_absolute():
        raise ValueError(
            f"project_path must be an absolute path. Received: {project_path}"
        )

    # only check a few conf keys that are known to specify a path string as value
    conf_keys_with_filepath = ("filename", "filepath", "path")

    for conf_key, conf_value in conf_dictionary.items():
        # if the conf_value is another dictionary, absolutify its paths first.
        if isinstance(conf_value, dict):
            conf_dictionary[conf_key] = _convert_paths_to_absolute_posix(
                project_path, conf_value
            )
            continue

        # if the conf_value is not a dictionary nor a string, skip
        if not isinstance(conf_value, str):
            continue

        # if the conf_value is a string but the conf_key isn't one associated with filepath, skip
        if conf_key not in conf_keys_with_filepath:
            continue

        if _is_relative_path(conf_value):
            # Absolute local path should be in POSIX format
            conf_value_absolute_path = (project_path / conf_value).as_posix()
            conf_dictionary[conf_key] = conf_value_absolute_path
        elif PureWindowsPath(conf_value).drive:
            # Convert absolute Windows path to POSIX format
            conf_dictionary[conf_key] = PureWindowsPath(conf_value).as_posix()

    return conf_dictionary


def _validate_transcoded_datasets(catalog: CatalogProtocol) -> None:
    """Validates transcoded datasets are correctly named

    Args:
        catalog: The catalog object containing the datasets to be
            validated.

    Raises:
        ValueError: If a dataset name does not conform to the expected
            transcoding naming conventions,a ValueError is raised by the
            `_transcode_split` function.

    """
    for dataset_name in catalog:
        _transcode_split(dataset_name)


def _expand_full_path(project_path: str | Path) -> Path:
    return Path(project_path).expanduser().resolve()


def _update_nested_dict(old_dict: dict[Any, Any], new_dict: dict[Any, Any]) -> None:
    """Update a nested dict with values of new_dict.
    Args:
        old_dict: dict to be updated
        new_dict: dict to use for updating old_dict
    """
    for key, value in new_dict.items():
        if key not in old_dict:
            old_dict[key] = value
        elif isinstance(old_dict[key], dict) and isinstance(value, dict):
            _update_nested_dict(old_dict[key], value)
        else:
            old_dict[key] = value


@define(slots=False)  # Enable setting new attributes to `KedroContext`
class KedroContext:
    """``KedroContext`` is the base class which holds the configuration and
    Kedro's main functionality.

    Create a context object by providing the root of a Kedro project and
    the environment configuration subfolders (see ``kedro.config.OmegaConfigLoader``)
    Raises:
        KedroContextError: If there is a mismatch
            between Kedro project version and package version.
    Args:
        project_path: Project path to define the context for.
        config_loader: Kedro's ``OmegaConfigLoader`` for loading the configuration files.
        env: Optional argument for configuration default environment to be used
            for running the pipeline. If not specified, it defaults to "local".
        package_name: Package name for the Kedro project the context is
            created for.
        hook_manager: The ``PluginManager`` to activate hooks, supplied by the session.
        runtime_params: Optional dictionary containing runtime project parameters.
            If specified, will update (and therefore take precedence over)
            the parameters retrieved from the project configuration.

    """

    project_path: Path = field(init=True, converter=_expand_full_path)
    config_loader: AbstractConfigLoader = field(init=True)
    env: str | None = field(init=True)
    _package_name: str = field(init=True)
    _hook_manager: PluginManager = field(init=True)
    _runtime_params: dict[str, Any] | None = field(
        init=True, default=None, converter=deepcopy
    )

    @property
    def catalog(self) -> CatalogProtocol:
        """Read-only property referring to Kedro's catalog` for this context.

        Returns:
            catalog defined in `catalog.yml`.
        Raises:
            KedroContextError: Incorrect catalog registered for the project.

        """
        return self._get_catalog()

    @property
    def params(self) -> dict[str, Any]:
        """Read-only property referring to Kedro's parameters for this context.

        Returns:
            Parameters defined in `parameters.yml` with the addition of any
                extra parameters passed at initialization.
        """
        try:
            params = self.config_loader["parameters"]
        except MissingConfigException as exc:
            warn(f"Parameters not found in your Kedro project config.\n{exc!s}")
            params = {}
        _update_nested_dict(params, self._runtime_params or {})
        return params  # type: ignore

    def _get_catalog(
        self,
        catalog_class: type = DataCatalog,
        save_version: str | None = None,
        load_versions: dict[str, str] | None = None,
    ) -> CatalogProtocol:
        """A hook for changing the creation of a catalog instance.

        Returns:
            catalog defined in `catalog.yml`.
        Raises:
            KedroContextError: Incorrect catalog registered for the project.

        """
        # '**/catalog*' reads modular pipeline configs
        conf_catalog = self.config_loader["catalog"]
        # turn relative paths in conf_catalog into absolute paths
        # before initializing the catalog
        conf_catalog = _convert_paths_to_absolute_posix(
            project_path=self.project_path, conf_dictionary=conf_catalog
        )
        conf_creds = self._get_config_credentials()

        if catalog_class is DataCatalog:
            catalog_class = compose_classes(catalog_class, CatalogCommandsMixin)

        catalog: CatalogProtocol = catalog_class.from_config(  # type: ignore[attr-defined]
            catalog=conf_catalog,
            credentials=conf_creds,
            load_versions=load_versions,
            save_version=save_version,
        )

        parameters = self._get_parameters()

        # Add parameters data to catalog.
        for param_name, param_value in parameters.items():
            catalog[param_name] = param_value

        _validate_transcoded_datasets(catalog)

        self._hook_manager.hook.after_catalog_created(
            catalog=catalog,
            conf_catalog=conf_catalog,
            conf_creds=conf_creds,
            parameters=parameters,
            save_version=save_version,
            load_versions=load_versions,
        )
        return catalog

    def _get_parameters(self) -> dict[str, Any]:
        """Returns a dictionary with data to be added in memory as `MemoryDataset`` instances.
        Keys represent parameter names and the values are parameter values."""
        params = self.params
        params_dict = {"parameters": params}

        def _add_param_to_params_dict(param_name: str, param_value: Any) -> None:
            """This recursively adds parameter paths that are defined in `parameters.yml`
            with the addition of any extra parameters passed at initialization to the `params_dict`,
            whenever `param_value` is a dictionary itself, so that users can
            specify specific nested parameters in their node inputs.

            Example:
            ``` python
            param_name = "a"
            param_value = {"b": 1}
            _add_param_to_params_dict(param_name, param_value)
            assert params_dict["params:a"] == {"b": 1}
            assert params_dict["params:a.b"] == 1
            ```
            """
            key = f"params:{param_name}"
            params_dict[key] = param_value
            if isinstance(param_value, dict):
                for key, val in param_value.items():
                    _add_param_to_params_dict(f"{param_name}.{key}", val)

        for param_name, param_value in params.items():
            _add_param_to_params_dict(param_name, param_value)

        return params_dict

    def _get_config_credentials(self) -> dict[str, Any]:
        """Getter for credentials specified in credentials directory."""
        try:
            conf_creds: dict[str, Any] = self.config_loader["credentials"]
        except MissingConfigException as exc:
            logging.getLogger(__name__).debug(
                "Credentials not found in your Kedro project config.\n %s", str(exc)
            )
            conf_creds = {}
        return conf_creds


class KedroContextError(Exception):
    """Error occurred when loading project and running context pipeline."""


def compose_classes(base: type, *mixins: type) -> type:
    """Injecting mixins dynamically"""
    t = (base, *mixins)
    name = f"{base.__name__}With{''.join(x.__name__ for x in mixins)}"
    return type(name, t, {})
